import asyncio
from inspect import isawaitable

from pytest import mark

from graphql.pyutils import is_awaitable


def describe_is_awaitable():
    def declines_the_none_value():
        assert not isawaitable(None)
        assert not is_awaitable(None)

    def declines_a_boolean_value():
        assert not isawaitable(True)
        assert not is_awaitable(True)

    def declines_an_int_value():
        assert not is_awaitable(42)

    def declines_a_string_value():
        assert not isawaitable("some_string")
        assert not is_awaitable("some_string")

    def declines_a_dict_value():
        assert not isawaitable({})
        assert not is_awaitable({})

    def declines_an_object_instance():
        assert not isawaitable(object())
        assert not is_awaitable(object())

    def declines_the_type_class():
        assert not isawaitable(type)
        assert not is_awaitable(type)

    def declines_a_lambda_function():
        assert not isawaitable(lambda: True)  # pragma: no cover
        assert not is_awaitable(lambda: True)  # pragma: no cover

    def declines_a_normal_function():
        def some_function():
            return True

        assert not isawaitable(some_function())
        assert not is_awaitable(some_function)

    def declines_a_normal_generator_function():
        def some_generator():
            yield True  # pragma: no cover

        assert not isawaitable(some_generator)
        assert not is_awaitable(some_generator)

    def declines_a_normal_generator_object():
        def some_generator():
            yield True  # pragma: no cover

        assert not isawaitable(some_generator())
        assert not is_awaitable(some_generator())

    def declines_a_coroutine_function():
        async def some_coroutine():
            return True  # pragma: no cover

        assert not isawaitable(some_coroutine)
        assert not is_awaitable(some_coroutine)

    @mark.asyncio
    @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
    async def recognizes_a_coroutine_object():
        async def some_coroutine():
            return False  # pragma: no cover

        assert isawaitable(some_coroutine())
        assert is_awaitable(some_coroutine())

    @mark.filterwarnings("ignore::Warning")  # Deprecation and Runtime
    def recognizes_an_old_style_coroutine():
        @asyncio.coroutine
        def some_old_style_coroutine():
            yield False  # pragma: no cover

        assert is_awaitable(some_old_style_coroutine())
        assert is_awaitable(some_old_style_coroutine())

    @mark.asyncio
    @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
    async def recognizes_a_future_object():
        async def some_coroutine():
            return False  # pragma: no cover

        some_future = asyncio.ensure_future(some_coroutine())

        assert is_awaitable(some_future)
        assert is_awaitable(some_future)

    @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
    def declines_an_async_generator():
        async def some_async_generator():
            yield True  # pragma: no cover

        assert not isawaitable(some_async_generator())
        assert not is_awaitable(some_async_generator())



from functools import partial

from graphql.validation import ScalarLeafsRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, ScalarLeafsRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_scalar_leafs():
    def valid_scalar_selection():
        assert_valid(
            """
            fragment scalarSelection on Dog {
              barks
            }
            """
        )

    def object_type_missing_selection():
        assert_errors(
            """
            query directQueryOnObjectWithoutSubFields {
              human
            }
            """,
            [
                {
                    "message": "Field 'human' of type 'Human'"
                    " must have a selection of subfields."
                    " Did you mean 'human { ... }'?",
                    "locations": [(3, 15)],
                },
            ],
        )

    def interface_type_missing_selection():
        assert_errors(
            """
            {
              human { pets }
            }
            """,
            [
                {
                    "message": "Field 'pets' of type '[Pet]'"
                    " must have a selection of subfields."
                    " Did you mean 'pets { ... }'?",
                    "locations": [(3, 23)],
                },
            ],
        )

    def valid_scalar_selection_with_args():
        assert_valid(
            """
            fragment scalarSelectionWithArgs on Dog {
              doesKnowCommand(dogCommand: SIT)
            }
            """
        )

    def scalar_selection_not_allowed_on_boolean():
        assert_errors(
            """
            fragment scalarSelectionsNotAllowedOnBoolean on Dog {
              barks { sinceWhen }
            }
            """,
            [
                {
                    "message": "Field 'barks' must not have a selection"
                    " since type 'Boolean' has no subfields.",
                    "locations": [(3, 21)],
                },
            ],
        )

    def scalar_selection_not_allowed_on_enum():
        assert_errors(
            """
            fragment scalarSelectionsNotAllowedOnEnum on Cat {
              furColor { inHexDec }
            }
            """,
            [
                {
                    "message": "Field 'furColor' must not have a selection"
                    " since type 'FurColor' has no subfields.",
                    "locations": [(3, 24)],
                },
            ],
        )

    def scalar_selection_not_allowed_with_args():
        assert_errors(
            """
            fragment scalarSelectionsNotAllowedWithArgs on Dog {
              doesKnowCommand(dogCommand: SIT) { sinceWhen }
            }
            """,
            [
                {
                    "message": "Field 'doesKnowCommand' must not have a selection"
                    " since type 'Boolean' has no subfields.",
                    "locations": [(3, 48)],
                },
            ],
        )

    def scalar_selection_not_allowed_with_directives():
        assert_errors(
            """
            fragment scalarSelectionsNotAllowedWithDirectives on Dog {
              name @include(if: true) { isAlsoHumanName }
            }
            """,
            [
                {
                    "message": "Field 'name' must not have a selection"
                    " since type 'String' has no subfields.",
                    "locations": [(3, 39)],
                },
            ],
        )

    def scalar_selection_not_allowed_with_directives_and_args():
        assert_errors(
            """
            fragment scalarSelectionsNotAllowedWithDirectivesAndArgs on Dog {
              doesKnowCommand(dogCommand: SIT) @include(if: true) { sinceWhen }
            }
            """,
            [
                {
                    "message": "Field 'doesKnowCommand' must not have a selection"
                    " since type 'Boolean' has no subfields.",
                    "locations": [(3, 67)],
                },
            ],
        )



from collections import defaultdict
from functools import cmp_to_key
from typing import Any, Dict, List, Union, cast

from ...type import (
    GraphQLAbstractType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLOutputType,
    GraphQLSchema,
    is_abstract_type,
    is_interface_type,
    is_object_type,
)
from ...error import GraphQLError
from ...language import FieldNode
from ...pyutils import did_you_mean, natural_comparison_key, suggestion_list
from . import ValidationRule

__all__ = ["FieldsOnCorrectTypeRule"]


class FieldsOnCorrectTypeRule(ValidationRule):
    """Fields on correct type

    A GraphQL document is only valid if all fields selected are defined by the parent
    type, or are an allowed meta field such as ``__typename``.

    See https://spec.graphql.org/draft/#sec-Field-Selections
    """

    def enter_field(self, node: FieldNode, *_args: Any) -> None:
        type_ = self.context.get_parent_type()
        if not type_:
            return
        field_def = self.context.get_field_def()
        if field_def:
            return
        # This field doesn't exist, lets look for suggestions.
        schema = self.context.schema
        field_name = node.name.value

        # First determine if there are any suggested types to condition on.
        suggestion = did_you_mean(
            get_suggested_type_names(schema, type_, field_name),
            "to use an inline fragment on",
        )

        # If there are no suggested types, then perhaps this was a typo?
        if not suggestion:
            suggestion = did_you_mean(get_suggested_field_names(type_, field_name))

        # Report an error, including helpful suggestions.
        self.report_error(
            GraphQLError(
                f"Cannot query field '{field_name}' on type '{type_}'." + suggestion,
                node,
            )
        )


def get_suggested_type_names(
    schema: GraphQLSchema, type_: GraphQLOutputType, field_name: str
) -> List[str]:
    """
    Get a list of suggested type names.

    Go through all of the implementations of type, as well as the interfaces
    that they implement. If any of those types include the provided field,
    suggest them, sorted by how often the type is referenced.
    """
    if not is_abstract_type(type_):
        # Must be an Object type, which does not have possible fields.
        return []

    type_ = cast(GraphQLAbstractType, type_)
    # Use a dict instead of a set for stable sorting when usage counts are the same
    suggested_types: Dict[Union[GraphQLObjectType, GraphQLInterfaceType], None] = {}
    usage_count: Dict[str, int] = defaultdict(int)
    for possible_type in schema.get_possible_types(type_):
        if field_name not in possible_type.fields:
            continue

        # This object type defines this field.
        suggested_types[possible_type] = None
        usage_count[possible_type.name] = 1

        for possible_interface in possible_type.interfaces:
            if field_name not in possible_interface.fields:
                continue

            # This interface type defines this field.
            suggested_types[possible_interface] = None
            usage_count[possible_interface.name] += 1

    def cmp(
        type_a: Union[GraphQLObjectType, GraphQLInterfaceType],
        type_b: Union[GraphQLObjectType, GraphQLInterfaceType],
    ) -> int:  # pragma: no cover
        # Suggest both interface and object types based on how common they are.
        usage_count_diff = usage_count[type_b.name] - usage_count[type_a.name]
        if usage_count_diff:
            return usage_count_diff

        # Suggest super types first followed by subtypes
        if is_interface_type(type_a) and schema.is_sub_type(
            cast(GraphQLInterfaceType, type_a), type_b
        ):
            return -1
        if is_interface_type(type_b) and schema.is_sub_type(
            cast(GraphQLInterfaceType, type_b), type_a
        ):
            return 1

        name_a = natural_comparison_key(type_a.name)
        name_b = natural_comparison_key(type_b.name)
        if name_a > name_b:
            return 1
        if name_a < name_b:
            return -1
        return 0

    return [type_.name for type_ in sorted(suggested_types, key=cmp_to_key(cmp))]


def get_suggested_field_names(type_: GraphQLOutputType, field_name: str) -> List[str]:
    """Get a list of suggested field names.

    For the field name provided, determine if there are any similar field names that may
    be the result of a typo.
    """
    if is_object_type(type_) or is_interface_type(type_):
        possible_field_names = list(type_.fields)  # type: ignore
        return suggestion_list(field_name, possible_field_names)
    # Otherwise, must be a Union type, which does not define fields.
    return []



from typing import Any, Collection, List, Union, cast

from ...error import GraphQLError
from ...language import (
    is_type_definition_node,
    is_type_system_definition_node,
    is_type_system_extension_node,
    Node,
    NamedTypeNode,
    TypeDefinitionNode,
)
from ...type import introspection_types, specified_scalar_types
from ...pyutils import did_you_mean, suggestion_list
from . import ASTValidationRule, ValidationContext, SDLValidationContext

__all__ = ["KnownTypeNamesRule"]


class KnownTypeNamesRule(ASTValidationRule):
    """Known type names

    A GraphQL document is only valid if referenced types (specifically variable
    definitions and fragment conditions) are defined by the type schema.

    See https://spec.graphql.org/draft/#sec-Fragment-Spread-Type-Existence
    """

    def __init__(self, context: Union[ValidationContext, SDLValidationContext]):
        super().__init__(context)
        schema = context.schema
        self.existing_types_map = schema.type_map if schema else {}

        defined_types = []
        for def_ in context.document.definitions:
            if is_type_definition_node(def_):
                def_ = cast(TypeDefinitionNode, def_)
                defined_types.append(def_.name.value)
        self.defined_types = set(defined_types)

        self.type_names = list(self.existing_types_map) + defined_types

    def enter_named_type(
        self,
        node: NamedTypeNode,
        _key: Any,
        parent: Node,
        _path: Any,
        ancestors: List[Node],
    ) -> None:
        type_name = node.name.value
        if (
            type_name not in self.existing_types_map
            and type_name not in self.defined_types
        ):
            try:
                definition_node = ancestors[2]
            except IndexError:
                definition_node = parent
            is_sdl = is_sdl_node(definition_node)
            if is_sdl and type_name in standard_type_names:
                return

            suggested_types = suggestion_list(
                type_name,
                list(standard_type_names) + self.type_names
                if is_sdl
                else self.type_names,
            )
            self.report_error(
                GraphQLError(
                    f"Unknown type '{type_name}'." + did_you_mean(suggested_types),
                    node,
                )
            )


standard_type_names = set(specified_scalar_types).union(introspection_types)


def is_sdl_node(value: Union[Node, Collection[Node], None]) -> bool:
    return (
        value is not None
        and not isinstance(value, list)
        and (
            is_type_system_definition_node(cast(Node, value))
            or is_type_system_extension_node(cast(Node, value))
        )
    )



from functools import partial
from typing import Callable, List, Tuple

from graphql.utilities import build_schema
from graphql.validation import NoDeprecatedCustomRule

from .harness import assert_validation_errors


def build_assertions(
    sdl_str: str,
) -> Tuple[Callable[[str], None], Callable[[str, List], None]]:
    schema = build_schema(sdl_str)
    assert_errors = partial(
        assert_validation_errors, NoDeprecatedCustomRule, schema=schema
    )
    assert_valid = partial(assert_errors, errors=[])
    return (
        assert_valid,
        assert_errors,
    )  # type: ignore


def describe_validate_no_deprecated():
    def describe_no_deprecated_fields():
        _assert_valid, _assert_errors = build_assertions(
            """
            type Query {
              normalField: String
              deprecatedField: String @deprecated(reason: "Some field reason.")
            }
            """
        )

        def ignores_fields_and_enum_values_that_are_not_deprecated():
            _assert_valid(
                """
                {
                  normalField
                }
                """
            )

        def ignores_unknown_fields():
            _assert_valid(
                """
                {
                  unknownField
                }

                fragment UnknownFragment on UnknownType {
                  deprecatedField
                }
                """
            )

        def reports_error_when_a_deprecated_field_is_selected():
            message = (
                "The field Query.deprecatedField is deprecated. Some field reason."
            )
            _assert_errors(
                """
                {
                  deprecatedField
                }

                fragment QueryFragment on Query {
                  deprecatedField
                }
                """,
                [
                    {
                        "message": message,
                        "locations": [(3, 19)],
                    },
                    {
                        "message": message,
                        "locations": [(7, 19)],
                    },
                ],
            )

    def describe_no_deprecated_arguments_on_fields():
        _assert_valid, _assert_errors = build_assertions(
            """
            type Query {
              someField(
                normalArg: String,
                deprecatedArg: String @deprecated(reason: "Some arg reason."),
              ): String
            }
            """
        )

        def ignores_arguments_that_are_not_deprecated():
            _assert_valid(
                """
                {
                  normalField(normalArg: "")
                }
                """
            )

        def ignores_unknown_arguments():
            _assert_valid(
                """
                {
                  someField(unknownArg: "")
                  unknownField(deprecatedArg: "")
                }
                """
            )

        def reports_error_when_a_deprecated_argument_is_used():
            _assert_errors(
                """
                {
                    someField(deprecatedArg: "")
                }
                """,
                [
                    {
                        "message": "Field 'Query.someField' argument 'deprecatedArg'"
                        " is deprecated. Some arg reason.",
                        "locations": [(3, 31)],
                    }
                ],
            )

    def describe_no_deprecated_arguments_on_directives():
        _assert_valid, _assert_errors = build_assertions(
            """
            type Query {
              someField: String
            }

            directive @someDirective(
              normalArg: String,
              deprecatedArg: String @deprecated(reason: "Some arg reason."),
            ) on FIELD
            """
        )

        def ignores_arguments_that_are_not_deprecated():
            _assert_valid(
                """
                {
                  someField @someDirective(normalArg: "")
                }
                """
            )

        def ignores_unknown_arguments():
            _assert_valid(
                """
                {
                  someField @someDirective(unknownArg: "")
                  someField @unknownDirective(deprecatedArg: "")
                }
                """
            )

        def reports_error_when_a_deprecated_argument_is_used():
            _assert_errors(
                """
                {
                  someField @someDirective(deprecatedArg: "")
                }
                """,
                [
                    {
                        "message": "Directive '@someDirective' argument 'deprecatedArg'"
                        " is deprecated. Some arg reason.",
                        "locations": [(3, 44)],
                    }
                ],
            )

    def describe_no_deprecated_input_fields():
        _assert_valid, _assert_errors = build_assertions(
            """
            input InputType {
              normalField: String
              deprecatedField: String @deprecated(reason: "Some input field reason.")
            }

            type Query {
              someField(someArg: InputType): String
            }

            directive @someDirective(someArg: InputType) on FIELD
            """
        )

        def ignores_input_fields_that_are_not_deprecated():
            _assert_valid(
                """
                {
                  someField(
                    someArg: { normalField: "" }
                  ) @someDirective(someArg: { normalField: "" })
                }
                """
            )

        def ignores_unknown_input_fields():
            _assert_valid(
                """
                {
                  someField(
                    someArg: { unknownField: "" }
                  )

                  someField(
                    unknownArg: { unknownField: "" }
                  )

                  unknownField(
                    unknownArg: { unknownField: "" }
                  )
                }
                """
            )

        def reports_error_when_a_deprecated_input_field_is_used():
            message = (
                "The input field InputType.deprecatedField is deprecated."
                " Some input field reason."
            )

            _assert_errors(
                """
                {
                  someField(
                    someArg: { deprecatedField: "" }
                  ) @someDirective(someArg: { deprecatedField: "" })
                }
                """,
                [
                    {"message": message, "locations": [(4, 32)]},
                    {"message": message, "locations": [(5, 47)]},
                ],
            )

    def describe_no_deprecated_enum_values():
        _assert_valid, _assert_errors = build_assertions(
            """
            enum EnumType {
              NORMAL_VALUE
              DEPRECATED_VALUE @deprecated(reason: "Some enum reason.")
            }

            type Query {
              someField(enumArg: EnumType): String
            }
            """
        )

        def ignores_enum_values_that_are_not_deprecated():
            _assert_valid(
                """
                {
                  normalField(enumArg: NORMAL_VALUE)
                }
                """
            )

        def ignores_unknown_enum_values():
            _assert_valid(
                """
                query (
                  $unknownValue: EnumType = UNKNOWN_VALUE
                  $unknownType: UnknownType = UNKNOWN_VALUE
                ) {
                  someField(enumArg: UNKNOWN_VALUE)
                  someField(unknownArg: UNKNOWN_VALUE)
                  unknownField(unknownArg: UNKNOWN_VALUE)
                }

                fragment SomeFragment on Query {
                  someField(enumArg: UNKNOWN_VALUE)
                }
                """
            )

        def reports_error_when_a_deprecated_enum_value_is_used():
            message = (
                "The enum value 'EnumType.DEPRECATED_VALUE' is deprecated."
                " Some enum reason."
            )
            _assert_errors(
                """
                query (
                  $variable: EnumType = DEPRECATED_VALUE
                ) {
                  someField(enumArg: DEPRECATED_VALUE)
                }
                """,
                [
                    {
                        "message": message,
                        "locations": [(3, 41)],
                    },
                    {
                        "message": message,
                        "locations": [(5, 38)],
                    },
                ],
            )



from graphql.type import (
    GraphQLField,
    GraphQLFloat,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLOutputType,
    GraphQLSchema,
    GraphQLString,
    GraphQLUnionType,
)
from graphql.utilities import is_equal_type, is_type_sub_type_of


def describe_type_comparators():
    def describe_is_equal_type():
        def same_references_are_equal():
            assert is_equal_type(GraphQLString, GraphQLString) is True

        def int_and_float_are_not_equal():
            assert is_equal_type(GraphQLInt, GraphQLFloat) is False

        def lists_of_same_type_are_equal():
            assert (
                is_equal_type(GraphQLList(GraphQLInt), GraphQLList(GraphQLInt)) is True
            )

        def lists_is_not_equal_to_item():
            assert is_equal_type(GraphQLList(GraphQLInt), GraphQLInt) is False

        def nonnull_of_same_type_are_equal():
            assert (
                is_equal_type(GraphQLNonNull(GraphQLInt), GraphQLNonNull(GraphQLInt))
                is True
            )

        def nonnull_is_not_equal_to_nullable():
            assert is_equal_type(GraphQLNonNull(GraphQLInt), GraphQLInt) is False

    def describe_is_type_sub_type_of():
        def _test_schema(field_type: GraphQLOutputType = GraphQLString):
            return GraphQLSchema(
                query=GraphQLObjectType("Query", {"field": GraphQLField(field_type)})
            )

        def same_reference_is_subtype():
            assert (
                is_type_sub_type_of(_test_schema(), GraphQLString, GraphQLString)
                is True
            )

        def int_is_not_subtype_of_float():
            assert (
                is_type_sub_type_of(_test_schema(), GraphQLInt, GraphQLFloat) is False
            )

        def non_null_is_subtype_of_nullable():
            assert (
                is_type_sub_type_of(
                    _test_schema(), GraphQLNonNull(GraphQLInt), GraphQLInt
                )
                is True
            )

        def nullable_is_not_subtype_of_non_null():
            assert (
                is_type_sub_type_of(
                    _test_schema(), GraphQLInt, GraphQLNonNull(GraphQLInt)
                )
                is False
            )

        def item_is_not_subtype_of_list():
            assert not is_type_sub_type_of(
                _test_schema(), GraphQLInt, GraphQLList(GraphQLInt)
            )

        def list_is_not_subtype_of_item():
            assert not is_type_sub_type_of(
                _test_schema(), GraphQLList(GraphQLInt), GraphQLInt
            )

        def member_is_subtype_of_union():
            member = GraphQLObjectType("Object", {"field": GraphQLField(GraphQLString)})
            union = GraphQLUnionType("Union", [member])
            schema = _test_schema(union)
            assert is_type_sub_type_of(schema, member, union)

        def implementing_object_is_subtype_of_interface():
            iface = GraphQLInterfaceType(
                "Interface", {"field": GraphQLField(GraphQLString)}
            )
            impl = GraphQLObjectType(
                "Object",
                {"field": GraphQLField(GraphQLString)},
                [iface],
            )
            schema = _test_schema(impl)
            assert is_type_sub_type_of(schema, impl, iface)

        def implementing_interface_is_subtype_of_interface():
            iface = GraphQLInterfaceType(
                "Interface", {"field": GraphQLField(GraphQLString)}
            )
            iface2 = GraphQLInterfaceType(
                "Interface2", {"field": GraphQLField(GraphQLString)}, [iface]
            )
            impl = GraphQLObjectType(
                "Object",
                {"field": GraphQLField(GraphQLString)},
                [iface2, iface],
            )
            schema = _test_schema(impl)
            assert is_type_sub_type_of(schema, iface2, iface)



from typing import Any, Collection, Dict, List, Optional, overload

from ..language import Node, OperationType
from ..pyutils import is_iterable


__all__ = ["ast_to_dict"]


@overload
def ast_to_dict(
    node: Node, locations: bool = False, cache: Optional[Dict[Node, Any]] = None
) -> Dict:
    ...


@overload
def ast_to_dict(
    node: Collection[Node],
    locations: bool = False,
    cache: Optional[Dict[Node, Any]] = None,
) -> List[Node]:
    ...


@overload
def ast_to_dict(
    node: OperationType,
    locations: bool = False,
    cache: Optional[Dict[Node, Any]] = None,
) -> str:
    ...


def ast_to_dict(
    node: Any, locations: bool = False, cache: Optional[Dict[Node, Any]] = None
) -> Any:
    """Convert a language AST to a nested Python dictionary.

    Set `location` to True in order to get the locations as well.
    """

    """Convert a node to a nested Python dictionary."""
    if isinstance(node, Node):
        if cache is None:
            cache = {}
        elif node in cache:
            return cache[node]
        cache[node] = res = {}
        res.update(
            {
                key: ast_to_dict(getattr(node, key), locations, cache)
                for key in ("kind",) + node.keys[1:]
            }
        )
        if locations:
            loc = node.loc
            if loc:
                res["loc"] = dict(start=loc.start, end=loc.end)
        return res
    if is_iterable(node):
        return [ast_to_dict(sub_node, locations, cache) for sub_node in node]
    if isinstance(node, OperationType):
        return node.value
    return node



from copy import copy, deepcopy
import weakref

from graphql.language import Location, Node, Source, Token, TokenKind
from graphql.pyutils import inspect


class SampleTestNode(Node):
    __slots__ = "alpha", "beta"

    alpha: int
    beta: int


def describe_token_class():
    def initializes():
        token = Token(
            kind=TokenKind.NAME,
            start=11,
            end=12,
            line=1,
            column=3,
            value="n",
        )
        assert token.kind == TokenKind.NAME
        assert token.start == 11
        assert token.end == 12
        assert token.line == 1
        assert token.column == 3
        assert token.prev is None
        assert token.value == "n"

    def can_stringify():
        token = Token(TokenKind.NAME, 1, 2, 1, 2, value="test")
        assert str(token) == "Name 'test'"
        assert token.desc == str(token)

    def has_representation_with_line_and_column():
        token = Token(TokenKind.NAME, 1, 2, 1, 2, value="test")
        assert repr(token) == "<Token Name 'test' 1:2>"
        assert inspect(token) == repr(token)

    def can_check_equality():
        token1 = Token(TokenKind.NAME, 1, 2, 1, 2, value="test")
        token2 = Token(TokenKind.NAME, 1, 2, 1, 2, value="test")
        assert token2 == token1
        assert not token2 != token1
        token3 = Token(TokenKind.NAME, 1, 2, 1, 2, value="text")
        assert token3 != token1
        token4 = Token(TokenKind.NAME, 1, 4, 1, 2, value="test")
        assert token4 != token1
        token5 = Token(TokenKind.NAME, 1, 2, 1, 4, value="test")
        assert token5 != token1

    def can_compare_with_string():
        token = Token(TokenKind.NAME, 1, 2, 1, 2, value="test")
        assert token == "Name 'test'"
        assert token != "Name 'foo'"

    def does_not_equal_incompatible_object():
        token = Token(TokenKind.NAME, 1, 2, 1, 2, value="test")
        assert token != {"Name": "test"}

    def can_hash():
        token1 = Token(TokenKind.NAME, 1, 2, 1, 2, value="hash")
        token2 = Token(TokenKind.NAME, 1, 2, 1, 2, value="hash")
        assert token2 == token1
        assert hash(token2) == hash(token1)
        token3 = Token(TokenKind.NAME, 1, 2, 1, 2, value="bash")
        assert token3 != token1
        assert hash(token3) != hash(token1)

    def can_copy():
        token1 = Token(TokenKind.NAME, 1, 2, 1, 2, value="copy")
        token2 = copy(token1)
        assert token2 == token1
        assert token2 is not token1


def describe_location_class():
    token1 = Token(TokenKind.NAME, 1, 2, 1, 2)
    token2 = Token(TokenKind.NAME, 2, 3, 1, 3)
    source = Source("source")

    def initializes():
        loc = Location(token1, token2, source)
        assert loc.start == token1.start
        assert loc.end == token2.end
        assert loc.start_token is token1
        assert loc.end_token is token2
        assert loc.source is source

    def can_stringify_with_start_and_end():
        loc = Location(token1, token2, source)
        assert str(loc) == "1:3"

    def has_representation_with_start_and_end():
        loc = Location(token1, token2, source)
        assert repr(loc) == "<Location 1:3>"
        assert inspect(loc) == repr(loc)

    def can_check_equality():
        loc1 = Location(token1, token2, source)
        loc2 = Location(token1, token2, source)
        assert loc2 == loc1
        loc3 = Location(token1, token1, source)
        assert loc3 != loc1
        loc4 = Location(token2, token2, source)
        assert loc4 != loc1
        assert loc4 != loc3

    def can_check_equality_with_tuple_or_list():
        loc = Location(token1, token2, source)
        assert loc == (1, 3)
        assert loc == [1, 3]
        assert not loc != (1, 3)
        assert not loc != [1, 3]
        assert loc != (1, 2)
        assert loc != [2, 3]

    def does_not_equal_incompatible_object():
        loc = Location(token1, token2, source)
        assert not loc == (1, 2, 3)
        assert loc != (1, 2, 3)
        assert not loc == {1: 2}
        assert loc != {1: 2}

    def can_hash():
        loc1 = Location(token1, token2, source)
        loc2 = Location(token1, token2, source)
        assert loc2 == loc1
        assert hash(loc2) == hash(loc1)
        loc3 = Location(token1, token1, source)
        assert loc3 != loc1
        assert hash(loc3) != hash(loc1)
        loc4 = Location(token2, token2, source)
        assert loc4 != loc1
        assert hash(loc4) != hash(loc1)
        assert hash(loc4) != hash(loc3)


def describe_node_class():
    def initializes_with_keywords():
        node = SampleTestNode(alpha=1, beta=2, loc=0)
        assert node.alpha == 1
        assert node.beta == 2
        assert node.loc == 0
        node = SampleTestNode(alpha=1, loc=None)
        assert node.loc is None
        assert node.alpha == 1
        assert node.beta is None
        node = SampleTestNode(alpha=1, beta=2, gamma=3)
        assert node.alpha == 1
        assert node.beta == 2
        assert not hasattr(node, "gamma")

    def has_representation_with_loc():
        node = SampleTestNode(alpha=1, beta=2)
        assert repr(node) == "SampleTestNode"
        node = SampleTestNode(alpha=1, beta=2, loc=3)
        assert repr(node) == "SampleTestNode at 3"

    def can_check_equality():
        node = SampleTestNode(alpha=1, beta=2)
        node2 = SampleTestNode(alpha=1, beta=2)
        assert node2 == node
        assert not node2 != node
        node2 = SampleTestNode(alpha=1, beta=1)
        assert node2 != node
        node3 = Node(alpha=1, beta=2)
        assert node3 != node

    def can_hash():
        node = SampleTestNode(alpha=1, beta=2)
        node2 = SampleTestNode(alpha=1, beta=2)
        assert node == node2
        assert node2 is not node
        assert hash(node2) == hash(node)
        node3 = SampleTestNode(alpha=1, beta=3)
        assert node3 != node
        assert hash(node3) != hash(node)

    def caches_are_hashed():
        node = SampleTestNode(alpha=1)
        assert not hasattr(node, "_hash")
        hash1 = hash(node)
        assert hasattr(node, "_hash")
        assert hash1 == getattr(node, "_hash")
        node.alpha = 2
        assert not hasattr(node, "_hash")
        hash2 = hash(node)
        assert hash2 != hash1
        assert hasattr(node, "_hash")
        assert hash2 == getattr(node, "_hash")

    def can_create_weak_reference():
        node = SampleTestNode(alpha=1, beta=2)
        ref = weakref.ref(node)
        assert ref() is node

    def can_create_custom_attribute():
        node = SampleTestNode(alpha=1, beta=2)
        node.gamma = 3
        assert node.gamma == 3  # type: ignore

    def can_create_shallow_copy():
        node = SampleTestNode(alpha=1, beta=2)
        node2 = copy(node)
        assert node2 is not node
        assert node2 == node

    def shallow_copy_is_really_shallow():
        node = SampleTestNode(alpha=1, beta=2)
        node2 = SampleTestNode(alpha=node, beta=node)
        node3 = copy(node2)
        assert node3 is not node2
        assert node3 == node2
        assert node3.alpha is node2.alpha
        assert node3.beta is node2.beta

    def can_create_deep_copy():
        alpha = SampleTestNode(alpha=1, beta=2)
        beta = SampleTestNode(alpha=3, beta=4)
        node = SampleTestNode(alpha=alpha, beta=beta)
        node2 = deepcopy(node)
        assert node2 is not node
        assert node2 == node
        assert node2.alpha == alpha
        assert node2.alpha is not alpha
        assert node2.alpha == alpha
        assert node2.beta is not beta

    def provides_snake_cased_kind_as_class_attribute():
        assert SampleTestNode.kind == "sample_test"

    def provides_proper_kind_if_class_does_not_end_with_node():
        class Foo(Node):
            pass

        assert Foo.kind == "foo"

    def provides_keys_as_class_attribute():
        assert SampleTestNode.keys == ("loc", "alpha", "beta")

    def can_can_convert_to_dict():
        node = SampleTestNode(alpha=1, beta=2)
        res = node.to_dict()
        assert node.to_dict(locations=True) == res
        assert res == {"kind": "sample_test", "alpha": 1, "beta": 2}
        assert list(res) == ["kind", "alpha", "beta"]

    def can_can_convert_to_dict_with_locations():
        token = Token(
            kind=TokenKind.NAME,
            start=1,
            end=3,
            line=1,
            column=1,
            value="foo",
        )
        loc = Location(token, token, Source("foo"))
        node = SampleTestNode(alpha=1, beta=2, loc=loc)
        res = node.to_dict(locations=True)
        assert res == {
            "kind": "sample_test",
            "alpha": 1,
            "beta": 2,
            "loc": {"start": 1, "end": 3},
        }
        assert list(res) == ["kind", "alpha", "beta", "loc"]
        assert list(res["loc"]) == ["start", "end"]



from typing import Union

from pytest import raises

from graphql import graphql_sync
from graphql.language import parse, print_ast
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLFloat,
    GraphQLID,
    GraphQLInputField,
    GraphQLInt,
    GraphQLNamedType,
    GraphQLSchema,
    GraphQLString,
    assert_directive,
    assert_enum_type,
    assert_input_object_type,
    assert_interface_type,
    assert_object_type,
    assert_scalar_type,
    assert_union_type,
    validate_schema,
)
from graphql.utilities import (
    build_schema,
    concat_ast,
    extend_schema,
    print_schema,
)

from ..utils import dedent

TypeWithAstNode = Union[
    GraphQLArgument,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLNamedType,
    GraphQLSchema,
]

TypeWithExtensionAstNodes = Union[
    GraphQLNamedType,
    GraphQLSchema,
]


def expect_extension_ast_nodes(obj: TypeWithExtensionAstNodes, expected: str) -> None:
    assert obj is not None and obj.extension_ast_nodes is not None
    assert "\n\n".join(print_ast(node) for node in obj.extension_ast_nodes) == expected


def expect_ast_node(obj: TypeWithAstNode, expected: str) -> None:
    assert obj is not None and obj.ast_node is not None
    assert print_ast(obj.ast_node) == expected


def expect_schema_changes(
    schema: GraphQLSchema, extended_schema: GraphQLSchema, expected: str
) -> None:
    schema_definitions = {
        print_ast(node) for node in parse(print_schema(schema)).definitions
    }
    assert (
        "\n\n".join(
            schema_def
            for schema_def in (
                print_ast(node)
                for node in parse(print_schema(extended_schema)).definitions
            )
            if schema_def not in schema_definitions
        )
        == expected
    )


def describe_extend_schema():
    def returns_the_original_schema_when_there_are_no_type_definitions():
        schema = build_schema("type Query")
        extended_schema = extend_schema(schema, parse("{ field }"))
        assert extended_schema == schema

    def can_be_used_for_limited_execution():
        schema = build_schema("type Query")
        extend_ast = parse(
            """
            extend type Query {
              newField: String
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        result = graphql_sync(
            schema=extended_schema, source="{ newField }", root_value={"newField": 123}
        )
        assert result == ({"newField": "123"}, None)

    def extends_objects_by_adding_new_fields():
        schema = build_schema(
            '''
            type Query {
              someObject: SomeObject
            }

            type SomeObject implements AnotherInterface & SomeInterface {
              self: SomeObject
              tree: [SomeObject]!
              """Old field description."""
              oldField: String
            }

            interface SomeInterface {
              self: SomeInterface
            }

            interface AnotherInterface {
              self: SomeObject
            }
            '''
        )
        extension_sdl = dedent(
            '''
            extend type SomeObject {
              """New field description."""
              newField(arg: Boolean): String
            }
          '''
        )
        extended_schema = extend_schema(schema, parse(extension_sdl))

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                '''
                type SomeObject implements AnotherInterface & SomeInterface {
                  self: SomeObject
                  tree: [SomeObject]!
                  """Old field description."""
                  oldField: String
                  """New field description."""
                  newField(arg: Boolean): String
                }
                '''
            ),
        )

    def extends_objects_with_standard_type_fields():
        schema = build_schema("type Query")

        # Only String and Boolean are used by introspection types
        assert schema.get_type("Int") is None
        assert schema.get_type("Float") is None
        assert schema.get_type("String") is GraphQLString
        assert schema.get_type("Boolean") is GraphQLBoolean
        assert schema.get_type("ID") is None

        extend_ast = parse(
            """
            extend type Query {
              bool: Boolean
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        assert extended_schema.get_type("Int") is None
        assert extended_schema.get_type("Float") is None
        assert extended_schema.get_type("String") is GraphQLString
        assert extended_schema.get_type("Boolean") is GraphQLBoolean
        assert extended_schema.get_type("ID") is None

        extend_twice_ast = parse(
            """
            extend type Query {
              int: Int
              float: Float
              id: ID
            }
            """
        )
        extended_twice_schema = extend_schema(schema, extend_twice_ast)

        assert validate_schema(extended_twice_schema) == []
        assert extended_twice_schema.get_type("Int") is GraphQLInt
        assert extended_twice_schema.get_type("Float") is GraphQLFloat
        assert extended_twice_schema.get_type("String") is GraphQLString
        assert extended_twice_schema.get_type("Boolean") is GraphQLBoolean
        assert extended_twice_schema.get_type("ID") is GraphQLID

    def extends_enums_by_adding_new_values():
        schema = build_schema(
            '''
            type Query {
              someEnum(arg: SomeEnum): SomeEnum
            }

            directive @foo(arg: SomeEnum) on SCHEMA

            enum SomeEnum {
              """Old value description."""
              OLD_VALUE
            }
            '''
        )
        extend_ast = parse(
            '''
            extend enum SomeEnum {
              """New value description."""
              NEW_VALUE
            }
            '''
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                '''
                enum SomeEnum {
                  """Old value description."""
                  OLD_VALUE
                  """New value description."""
                  NEW_VALUE
                }
                '''
            ),
        )

    def extends_unions_by_adding_new_types():
        schema = build_schema(
            """
            type Query {
              someUnion: SomeUnion
            }

            union SomeUnion = Foo | Biz

            type Foo { foo: String }
            type Biz { biz: String }
            type Bar { bar: String }
            """
        )
        extend_ast = parse(
            """
            extend union SomeUnion = Bar
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
                union SomeUnion = Foo | Biz | Bar
                """
            ),
        )

    def allows_extension_of_union_by_adding_itself():
        schema = build_schema(
            """
            union SomeUnion
            """
        )
        extend_ast = parse(
            """
            extend union SomeUnion = SomeUnion
            """
        )
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            extend_schema(schema, extend_ast)
        assert str(exc_info.value) == (
            "SomeUnion types must be specified"
            " as a collection of GraphQLObjectType instances."
        )

    def extends_inputs_by_adding_new_fields():
        schema = build_schema(
            '''
            type Query {
              someInput(arg: SomeInput): String
            }

            directive @foo(arg: SomeInput) on SCHEMA

            input SomeInput {
              """Old field description."""
              oldField: String
            }
            '''
        )
        extend_ast = parse(
            '''
            extend input SomeInput {
              """New field description."""
              newField: String
            }
            '''
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                '''
                input SomeInput {
                  """Old field description."""
                  oldField: String
                  """New field description."""
                  newField: String
                }
                '''
            ),
        )

    def extends_scalars_by_adding_new_directives():
        schema = build_schema(
            """
            type Query {
              someScalar(arg: SomeScalar): SomeScalar
            }

            directive @foo(arg: SomeScalar) on SCALAR

            input FooInput {
              foo: SomeScalar
            }

            scalar SomeScalar
            """
        )
        extension_sdl = dedent(
            """
            extend scalar SomeScalar @foo
            """
        )
        extended_schema = extend_schema(schema, parse(extension_sdl))
        some_scalar = assert_scalar_type(extended_schema.get_type("SomeScalar"))

        assert validate_schema(extended_schema) == []
        expect_extension_ast_nodes(some_scalar, extension_sdl)

    def extends_scalars_by_adding_specified_by_directive():
        schema = build_schema(
            """
            type Query {
              foo: Foo
            }

            scalar Foo

            directive @foo on SCALAR
            """
        )
        extension_sdl = dedent(
            """
            extend scalar Foo @foo

            extend scalar Foo @specifiedBy(url: "https://example.com/foo_spec")
            """
        )

        extended_schema = extend_schema(schema, parse(extension_sdl))
        foo = assert_scalar_type(extended_schema.get_type("Foo"))

        assert foo.specified_by_url == "https://example.com/foo_spec"

        assert validate_schema(extended_schema) == []
        expect_extension_ast_nodes(foo, extension_sdl)

    def correctly_assigns_ast_nodes_to_new_and_extended_types():
        schema = build_schema(
            """
            type Query

            scalar SomeScalar
            enum SomeEnum
            union SomeUnion
            input SomeInput
            type SomeObject
            interface SomeInterface

            directive @foo on SCALAR
            """
        )
        first_extension_ast = parse(
            """
            extend type Query {
              newField(testArg: TestInput): TestEnum
            }

            extend scalar SomeScalar @foo

            extend enum SomeEnum {
              NEW_VALUE
            }

            extend union SomeUnion = SomeObject

            extend input SomeInput {
              newField: String
            }

            extend interface SomeInterface {
              newField: String
            }

            enum TestEnum {
              TEST_VALUE
            }

            input TestInput {
              testInputField: TestEnum
            }
            """
        )
        extended_schema = extend_schema(schema, first_extension_ast)

        second_extension_ast = parse(
            """
            extend type Query {
              oneMoreNewField: TestUnion
            }

            extend scalar SomeScalar @test

            extend enum SomeEnum {
              ONE_MORE_NEW_VALUE
            }

            extend union SomeUnion = TestType

            extend input SomeInput {
              oneMoreNewField: String
            }

            extend interface SomeInterface {
              oneMoreNewField: String
            }

            union TestUnion = TestType

            interface TestInterface {
              interfaceField: String
            }

            type TestType implements TestInterface {
              interfaceField: String
            }

            directive @test(arg: Int) repeatable on FIELD | SCALAR
            """
        )
        extended_twice_schema = extend_schema(extended_schema, second_extension_ast)

        extend_in_one_go_schema = extend_schema(
            schema, concat_ast([first_extension_ast, second_extension_ast])
        )
        assert print_schema(extend_in_one_go_schema) == print_schema(
            extended_twice_schema
        )

        query = assert_object_type(extended_twice_schema.get_type("Query"))
        some_enum = assert_enum_type(extended_twice_schema.get_type("SomeEnum"))
        some_union = assert_union_type(extended_twice_schema.get_type("SomeUnion"))
        some_scalar = assert_scalar_type(extended_twice_schema.get_type("SomeScalar"))
        some_input = assert_input_object_type(
            extended_twice_schema.get_type("SomeInput")
        )
        some_interface = assert_interface_type(
            extended_twice_schema.get_type("SomeInterface")
        )

        test_input = assert_input_object_type(
            extended_twice_schema.get_type("TestInput")
        )
        test_enum = assert_enum_type(extended_twice_schema.get_type("TestEnum"))
        test_union = assert_union_type(extended_twice_schema.get_type("TestUnion"))
        test_type = assert_object_type(extended_twice_schema.get_type("TestType"))
        test_interface = assert_interface_type(
            extended_twice_schema.get_type("TestInterface")
        )
        test_directive = assert_directive(extended_twice_schema.get_directive("test"))

        assert test_type.extension_ast_nodes == ()
        assert test_enum.extension_ast_nodes == ()
        assert test_union.extension_ast_nodes == ()
        assert test_input.extension_ast_nodes == ()
        assert test_interface.extension_ast_nodes == ()

        assert query.extension_ast_nodes
        assert len(query.extension_ast_nodes) == 2
        assert some_scalar.extension_ast_nodes
        assert len(some_scalar.extension_ast_nodes) == 2
        assert some_enum.extension_ast_nodes
        assert len(some_enum.extension_ast_nodes) == 2
        assert some_union.extension_ast_nodes
        assert len(some_union.extension_ast_nodes) == 2
        assert some_input.extension_ast_nodes
        assert len(some_input.extension_ast_nodes) == 2
        assert some_interface.extension_ast_nodes
        assert len(some_interface.extension_ast_nodes) == 2

        assert {
            test_input.ast_node,
            test_enum.ast_node,
            test_union.ast_node,
            test_interface.ast_node,
            test_type.ast_node,
            test_directive.ast_node,
            *query.extension_ast_nodes,
            *some_scalar.extension_ast_nodes,
            *some_enum.extension_ast_nodes,
            *some_union.extension_ast_nodes,
            *some_input.extension_ast_nodes,
            *some_interface.extension_ast_nodes,
        } == {*first_extension_ast.definitions, *second_extension_ast.definitions}

        new_field = query.fields["newField"]
        expect_ast_node(new_field, "newField(testArg: TestInput): TestEnum")
        expect_ast_node(new_field.args["testArg"], "testArg: TestInput")
        expect_ast_node(query.fields["oneMoreNewField"], "oneMoreNewField: TestUnion")

        expect_ast_node(some_enum.values["NEW_VALUE"], "NEW_VALUE")

        one_more_new_value = some_enum.values["ONE_MORE_NEW_VALUE"]
        expect_ast_node(one_more_new_value, "ONE_MORE_NEW_VALUE")
        expect_ast_node(some_input.fields["newField"], "newField: String")
        expect_ast_node(some_input.fields["oneMoreNewField"], "oneMoreNewField: String")
        expect_ast_node(some_interface.fields["newField"], "newField: String")
        expect_ast_node(
            some_interface.fields["oneMoreNewField"], "oneMoreNewField: String"
        )

        expect_ast_node(test_input.fields["testInputField"], "testInputField: TestEnum")

        expect_ast_node(test_enum.values["TEST_VALUE"], "TEST_VALUE")

        expect_ast_node(
            test_interface.fields["interfaceField"], "interfaceField: String"
        )
        expect_ast_node(test_type.fields["interfaceField"], "interfaceField: String")
        expect_ast_node(test_directive.args["arg"], "arg: Int")

    def builds_types_with_deprecated_fields_and_values():
        schema = GraphQLSchema()
        extend_ast = parse(
            """
            type SomeObject {
              deprecatedField: String @deprecated(reason: "not used anymore")
            }

            enum SomeEnum {
              DEPRECATED_VALUE @deprecated(reason: "do not use")
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        some_type = assert_object_type(extended_schema.get_type("SomeObject"))
        deprecated_field = some_type.fields["deprecatedField"]
        assert deprecated_field.deprecation_reason == "not used anymore"

        some_enum = assert_enum_type(extended_schema.get_type("SomeEnum"))
        deprecated_enum = some_enum.values["DEPRECATED_VALUE"]
        assert deprecated_enum.deprecation_reason == "do not use"

    def extends_objects_with_deprecated_fields():
        schema = build_schema("type SomeObject")
        extend_ast = parse(
            """
            extend type SomeObject {
              deprecatedField: String @deprecated(reason: "not used anymore")
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        some_type = assert_object_type(extended_schema.get_type("SomeObject"))
        deprecated_field = some_type.fields["deprecatedField"]
        assert deprecated_field.deprecation_reason == "not used anymore"

    def extend_enums_with_deprecated_values():
        schema = build_schema("enum SomeEnum")
        extend_ast = parse(
            """
            extend enum SomeEnum {
              DEPRECATED_VALUE @deprecated(reason: "do not use")
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        some_enum = assert_enum_type(extended_schema.get_type("SomeEnum"))
        deprecated_value = some_enum.values["DEPRECATED_VALUE"]
        assert deprecated_value.deprecation_reason == "do not use"

    def adds_new_unused_types():
        schema = build_schema(
            """
            type Query {
              dummy: String
            }
            """
        )
        extension_sdl = dedent(
            """
            type DummyUnionMember {
              someField: String
            }

            enum UnusedEnum {
              SOME_VALUE
            }

            input UnusedInput {
              someField: String
            }

            interface UnusedInterface {
              someField: String
            }

            type UnusedObject {
              someField: String
            }

            union UnusedUnion = DummyUnionMember
            """
        )
        extended_schema = extend_schema(schema, parse(extension_sdl))

        assert validate_schema(extended_schema) == []
        expect_schema_changes(schema, extended_schema, extension_sdl)

    def extends_objects_by_adding_new_fields_with_arguments():
        schema = build_schema(
            """
            type SomeObject

            type Query {
              someObject: SomeObject
            }
            """
        )
        extend_ast = parse(
            """
            input NewInputObj {
              field1: Int
              field2: [Float]
              field3: String!
            }

            extend type SomeObject {
              newField(arg1: String, arg2: NewInputObj!): String
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
            type SomeObject {
              newField(arg1: String, arg2: NewInputObj!): String
            }

            input NewInputObj {
              field1: Int
              field2: [Float]
              field3: String!
            }
            """
            ),
        )

    def extends_objects_by_adding_new_fields_with_existing_types():
        schema = build_schema(
            """
            type Query {
              someObject: SomeObject
            }

            type SomeObject
            enum SomeEnum { VALUE }
            """
        )
        extend_ast = parse(
            """
            extend type SomeObject {
              newField(arg1: SomeEnum!): SomeEnum
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
            type SomeObject {
              newField(arg1: SomeEnum!): SomeEnum
            }
            """
            ),
        )

    def extends_objects_by_adding_implemented_interfaces():
        schema = build_schema(
            """
            type Query {
              someObject: SomeObject
            }

            type SomeObject {
              foo: String
            }

            interface SomeInterface {
              foo: String
            }
            """
        )
        extend_ast = parse(
            """
            extend type SomeObject implements SomeInterface
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
            type SomeObject implements SomeInterface {
              foo: String
            }
            """
            ),
        )

    def extends_objects_by_including_new_types():
        schema = build_schema(
            """
            type Query {
              someObject: SomeObject
            }

            type SomeObject {
              oldField: String
            }
            """
        )
        new_types_sdl = """
            enum NewEnum {
              VALUE
            }

            interface NewInterface {
              baz: String
            }

            type NewObject implements NewInterface {
              baz: String
            }

            scalar NewScalar

            union NewUnion = NewObject
            """
        extend_ast = parse(
            new_types_sdl
            + """
            extend type SomeObject {
              newObject: NewObject
              newInterface: NewInterface
              newUnion: NewUnion
              newScalar: NewScalar
              newEnum: NewEnum
              newTree: [SomeObject]!
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
            type SomeObject {
              oldField: String
              newObject: NewObject
              newInterface: NewInterface
              newUnion: NewUnion
              newScalar: NewScalar
              newEnum: NewEnum
              newTree: [SomeObject]!
            }\n"""
                + new_types_sdl
            ),
        )

    def extends_objects_by_adding_implemented_new_interfaces():
        schema = build_schema(
            """
            type Query {
              someObject: SomeObject
            }

            type SomeObject implements OldInterface {
              oldField: String
            }

            interface OldInterface {
              oldField: String
            }
            """
        )
        extend_ast = parse(
            """
            extend type SomeObject implements NewInterface {
              newField: String
            }

            interface NewInterface {
              newField: String
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
            type SomeObject implements OldInterface & NewInterface {
              oldField: String
              newField: String
            }

            interface NewInterface {
              newField: String
            }
            """
            ),
        )

    def extends_different_types_multiple_times():
        schema = build_schema(
            """
            type Query {
              someScalar: SomeScalar
              someObject(someInput: SomeInput): SomeObject
              someInterface: SomeInterface
              someEnum: SomeEnum
              someUnion: SomeUnion
            }

            scalar SomeScalar

            type SomeObject implements SomeInterface {
              oldField: String
            }

            interface SomeInterface {
              oldField: String
            }

            enum SomeEnum {
              OLD_VALUE
            }

            union SomeUnion = SomeObject

            input SomeInput {
              oldField: String
            }
            """
        )
        new_types_sdl = dedent(
            """
            scalar NewScalar

            scalar AnotherNewScalar

            type NewObject {
              foo: String
            }

            type AnotherNewObject {
              foo: String
            }

            interface NewInterface {
              newField: String
            }

            interface AnotherNewInterface {
              anotherNewField: String
            }
            """
        )
        schema_with_new_types = extend_schema(schema, parse(new_types_sdl))
        expect_schema_changes(schema, schema_with_new_types, new_types_sdl)

        extend_ast = parse(
            """
            extend scalar SomeScalar @specifiedBy(url: "http://example.com/foo_spec")

            extend type SomeObject implements NewInterface {
                newField: String
            }

            extend type SomeObject implements AnotherNewInterface {
                anotherNewField: String
            }

            extend enum SomeEnum {
                NEW_VALUE
            }

            extend enum SomeEnum {
                ANOTHER_NEW_VALUE
            }

            extend union SomeUnion = NewObject

            extend union SomeUnion = AnotherNewObject

            extend input SomeInput {
                newField: String
            }

            extend input SomeInput {
                anotherNewField: String
            }
            """
        )
        extended_schema = extend_schema(schema_with_new_types, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
                scalar SomeScalar @specifiedBy(url: "http://example.com/foo_spec")

                type SomeObject implements SomeInterface & NewInterface & AnotherNewInterface {
                  oldField: String
                  newField: String
                  anotherNewField: String
                }

                enum SomeEnum {
                  OLD_VALUE
                  NEW_VALUE
                  ANOTHER_NEW_VALUE
                }

                union SomeUnion = SomeObject | NewObject | AnotherNewObject

                input SomeInput {
                  oldField: String
                  newField: String
                  anotherNewField: String
                }

                """  # noqa: E501
            )
            + "\n\n"
            + new_types_sdl,
        )

    def extends_interfaces_by_adding_new_fields():
        schema = build_schema(
            """
            interface SomeInterface {
              oldField: String
            }

            interface AnotherInterface implements SomeInterface {
              oldField: String
            }

            type SomeObject implements SomeInterface & AnotherInterface {
              oldField: String
            }

            type Query {
              someInterface: SomeInterface
            }
            """
        )
        extend_ast = parse(
            """
            extend interface SomeInterface {
              newField: String
            }

            extend interface AnotherInterface {
              newField: String
            }

            extend type SomeObject {
              newField: String
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
                interface SomeInterface {
                  oldField: String
                  newField: String
                }

                interface AnotherInterface implements SomeInterface {
                  oldField: String
                  newField: String
                }

                type SomeObject implements SomeInterface & AnotherInterface {
                  oldField: String
                  newField: String
                }
                """
            ),
        )

    def extends_interfaces_by_adding_new_implemented_interfaces():
        schema = build_schema(
            """
            interface SomeInterface {
              oldField: String
            }

            interface AnotherInterface implements SomeInterface {
              oldField: String
            }

            type SomeObject implements SomeInterface & AnotherInterface {
              oldField: String
            }

            type Query {
              someInterface: SomeInterface
            }
            """
        )
        extend_ast = parse(
            """
            interface NewInterface {
              newField: String
            }

            extend interface AnotherInterface implements NewInterface {
              newField: String
            }

            extend type SomeObject implements NewInterface {
              newField: String
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
                interface AnotherInterface implements SomeInterface & NewInterface {
                  oldField: String
                  newField: String
                }

                type SomeObject implements SomeInterface & AnotherInterface & NewInterface {
                  oldField: String
                  newField: String
                }

                interface NewInterface {
                  newField: String
                }
                """  # noqa: E501
            ),
        )

    def allows_extension_of_interface_with_missing_object_fields():
        schema = build_schema(
            """
            type Query {
              someInterface: SomeInterface
            }

            type SomeObject implements SomeInterface {
              oldField: SomeInterface
            }

            interface SomeInterface {
              oldField: SomeInterface
            }
            """
        )
        extend_ast = parse(
            """
            extend interface SomeInterface {
              newField: String
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema)
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
                interface SomeInterface {
                  oldField: SomeInterface
                  newField: String
                }
                """
            ),
        )

    def extends_interfaces_multiple_times():
        schema = build_schema(
            """
            type Query {
              someInterface: SomeInterface
            }

            interface SomeInterface {
              some: SomeInterface
            }
            """
        )
        extend_ast = parse(
            """
            extend interface SomeInterface {
              newFieldA: Int
            }

            extend interface SomeInterface {
              newFieldB(test: Boolean): String
            }
            """
        )
        extended_schema = extend_schema(schema, extend_ast)

        assert validate_schema(extended_schema) == []
        expect_schema_changes(
            schema,
            extended_schema,
            dedent(
                """
                interface SomeInterface {
                  some: SomeInterface
                  newFieldA: Int
                  newFieldB(test: Boolean): String
                }
                """
            ),
        )

    def may_extend_mutations_and_subscriptions():
        mutation_schema = build_schema(
            """
            type Query {
              queryField: String
            }

            type Mutation {
              mutationField: String
            }

            type Subscription {
              subscriptionField: String
            }
            """
        )
        ast = parse(
            """
            extend type Query {
              newQueryField: Int
            }

            extend type Mutation {
              newMutationField: Int
            }

            extend type Subscription {
              newSubscriptionField: Int
            }
            """
        )
        original_print = print_schema(mutation_schema)
        extended_schema = extend_schema(mutation_schema, ast)
        assert extended_schema != mutation_schema
        assert print_schema(mutation_schema) == original_print
        assert print_schema(extended_schema) == dedent(
            """
            type Query {
              queryField: String
              newQueryField: Int
            }

            type Mutation {
              mutationField: String
              newMutationField: Int
            }

            type Subscription {
              subscriptionField: String
              newSubscriptionField: Int
            }
            """
        )

    def may_extend_directives_with_new_directive():
        schema = build_schema(
            """
            type Query {
              foo: String
            }
            """
        )
        extension_sdl = dedent(
            '''
            """New directive."""
            directive @new(enable: Boolean!, tag: String) repeatable on QUERY | FIELD
            '''
        )
        extended_schema = extend_schema(schema, parse(extension_sdl))

        assert validate_schema(extended_schema) == []
        expect_schema_changes(schema, extended_schema, extension_sdl)

    def rejects_invalid_sdl():
        schema = GraphQLSchema()
        extend_ast = parse("extend schema @unknown")

        with raises(TypeError) as exc_info:
            extend_schema(schema, extend_ast)
        assert str(exc_info.value) == "Unknown directive '@unknown'."

    def allows_to_disable_sdl_validation():
        schema = GraphQLSchema()
        extend_ast = parse("extend schema @unknown")

        extend_schema(schema, extend_ast, assume_valid=True)
        extend_schema(schema, extend_ast, assume_valid_sdl=True)

    def throws_on_unknown_types():
        schema = GraphQLSchema()
        ast = parse(
            """
            type Query {
              unknown: UnknownType
            }
            """
        )
        with raises(TypeError) as exc_info:
            extend_schema(schema, ast, assume_valid_sdl=True)
        assert str(exc_info.value).endswith("Unknown type: 'UnknownType'.")

    def rejects_invalid_ast():
        schema = GraphQLSchema()

        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            extend_schema(schema, None)  # type: ignore
        assert str(exc_info.value) == "Must provide valid Document AST."

        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            extend_schema(schema, {})  # type: ignore
        assert str(exc_info.value) == "Must provide valid Document AST."

    def does_not_allow_replacing_a_default_directive():
        schema = GraphQLSchema()
        extend_ast = parse(
            """
            directive @include(if: Boolean!) on FIELD | FRAGMENT_SPREAD
            """
        )

        with raises(TypeError) as exc_info:
            extend_schema(schema, extend_ast)
        assert str(exc_info.value).startswith(
            "Directive '@include' already exists in the schema."
            " It cannot be redefined."
        )

    def does_not_allow_replacing_an_existing_enum_value():
        schema = build_schema(
            """
            enum SomeEnum {
              ONE
            }
            """
        )
        extend_ast = parse(
            """
            extend enum SomeEnum {
              ONE
            }
            """
        )

        with raises(TypeError) as exc_info:
            extend_schema(schema, extend_ast)
        assert str(exc_info.value).startswith(
            "Enum value 'SomeEnum.ONE' already exists in the schema."
            " It cannot also be defined in this type extension."
        )

    def describe_can_add_additional_root_operation_types():
        def does_not_automatically_include_common_root_type_names():
            schema = GraphQLSchema()
            extended_schema = extend_schema(schema, parse("type Mutation"))

            assert extended_schema.get_type("Mutation")
            assert extended_schema.mutation_type is None

        def adds_schema_definition_missing_in_the_original_schema():
            schema = build_schema(
                """
                directive @foo on SCHEMA
                type Foo
                """
            )
            assert schema.query_type is None

            extension_sdl = dedent(
                """
                schema @foo {
                  query: Foo
                }
                """
            )
            extended_schema = extend_schema(schema, parse(extension_sdl))

            query_type = assert_object_type(extended_schema.query_type)
            assert query_type.name == "Foo"
            expect_ast_node(extended_schema, extension_sdl)

        def adds_new_root_types_via_schema_extension():
            schema = build_schema(
                """
                type Query
                type MutationRoot
                """
            )
            extension_sdl = dedent(
                """
                extend schema {
                  mutation: MutationRoot
                }
                """
            )
            extended_schema = extend_schema(schema, parse(extension_sdl))

            mutation_type = assert_object_type(extended_schema.mutation_type)
            assert mutation_type.name == "MutationRoot"
            expect_extension_ast_nodes(extended_schema, extension_sdl)

        def adds_directive_via_schema_extension():
            schema = build_schema(
                """
                type Query

                directive @foo on SCHEMA
                """
            )
            extension_sdl = dedent(
                """
                extend schema @foo
                """
            )
            extended_schema = extend_schema(schema, parse(extension_sdl))

            expect_extension_ast_nodes(extended_schema, extension_sdl)

        def adds_multiple_new_root_types_via_schema_extension():
            schema = build_schema("type Query")
            extend_ast = parse(
                """
                extend schema {
                  mutation: Mutation
                  subscription: Subscription
                }

                type Mutation
                type Subscription
                """
            )
            extended_schema = extend_schema(schema, extend_ast)

            mutation_type = assert_object_type(extended_schema.mutation_type)
            assert mutation_type.name == "Mutation"

            subscription_type = assert_object_type(extended_schema.subscription_type)
            assert subscription_type.name == "Subscription"

        def applies_multiple_schema_extensions():
            schema = build_schema("type Query")
            extend_ast = parse(
                """
                extend schema {
                  mutation: Mutation
                }
                type Mutation

                extend schema {
                  subscription: Subscription
                }
                type Subscription
                """
            )
            extended_schema = extend_schema(schema, extend_ast)

            mutation_type = assert_object_type(extended_schema.mutation_type)
            assert mutation_type.name == "Mutation"

            subscription_type = assert_object_type(extended_schema.subscription_type)
            assert subscription_type.name == "Subscription"

        def schema_extension_ast_are_available_from_schema_object():
            schema = build_schema(
                """
                type Query

                directive @foo on SCHEMA
                """
            )
            extend_ast = parse(
                """
                extend schema {
                  mutation: Mutation
                }
                type Mutation

                extend schema {
                  subscription: Subscription
                }
                type Subscription
                """
            )
            extended_schema = extend_schema(schema, extend_ast)

            second_extend_ast = parse("extend schema @foo")
            extended_twice_schema = extend_schema(extended_schema, second_extend_ast)

            expect_extension_ast_nodes(
                extended_twice_schema,
                dedent(
                    """
                    extend schema {
                      mutation: Mutation
                    }

                    extend schema {
                      subscription: Subscription
                    }

                    extend schema @foo
                    """
                ),
            )



from typing import Any, Collection, Optional

from ..language.ast import Node, OperationType
from .block_string import print_block_string
from .print_string import print_string
from .visitor import visit, Visitor

__all__ = ["print_ast"]


MAX_LINE_LENGTH = 80

Strings = Collection[str]


class PrintedNode:
    """A union type for all nodes that have been processed by the printer."""

    alias: str
    arguments: Strings
    block: bool
    default_value: str
    definitions: Strings
    description: str
    directives: str
    fields: Strings
    interfaces: Strings
    locations: Strings
    name: str
    operation: OperationType
    operation_types: Strings
    repeatable: bool
    selection_set: str
    selections: Strings
    type: str
    type_condition: str
    types: Strings
    value: str
    values: Strings
    variable: str
    variable_definitions: Strings


def print_ast(ast: Node) -> str:
    """Convert an AST into a string.

    The conversion is done using a set of reasonable formatting rules.
    """
    return visit(ast, PrintAstVisitor())


class PrintAstVisitor(Visitor):
    @staticmethod
    def leave_name(node: PrintedNode, *_args: Any) -> str:
        return node.value

    @staticmethod
    def leave_variable(node: PrintedNode, *_args: Any) -> str:
        return f"${node.name}"

    # Document

    @staticmethod
    def leave_document(node: PrintedNode, *_args: Any) -> str:
        return join(node.definitions, "\n\n")

    @staticmethod
    def leave_operation_definition(node: PrintedNode, *_args: Any) -> str:
        var_defs = wrap("(", join(node.variable_definitions, ", "), ")")
        prefix = join(
            (
                node.operation.value,
                join((node.name, var_defs)),
                join(node.directives, " "),
            ),
            " ",
        )
        # Anonymous queries with no directives or variable definitions can use the
        # query short form.
        return ("" if prefix == "query" else prefix + " ") + node.selection_set

    @staticmethod
    def leave_variable_definition(node: PrintedNode, *_args: Any) -> str:
        return (
            f"{node.variable}: {node.type}"
            f"{wrap(' = ', node.default_value)}"
            f"{wrap(' ', join(node.directives, ' '))}"
        )

    @staticmethod
    def leave_selection_set(node: PrintedNode, *_args: Any) -> str:
        return block(node.selections)

    @staticmethod
    def leave_field(node: PrintedNode, *_args: Any) -> str:
        prefix = wrap("", node.alias, ": ") + node.name
        args_line = prefix + wrap("(", join(node.arguments, ", "), ")")

        if len(args_line) > MAX_LINE_LENGTH:
            args_line = prefix + wrap("(\n", indent(join(node.arguments, "\n")), "\n)")

        return join((args_line, join(node.directives, " "), node.selection_set), " ")

    @staticmethod
    def leave_argument(node: PrintedNode, *_args: Any) -> str:
        return f"{node.name}: {node.value}"

    # Fragments

    @staticmethod
    def leave_fragment_spread(node: PrintedNode, *_args: Any) -> str:
        return f"...{node.name}{wrap(' ', join(node.directives, ' '))}"

    @staticmethod
    def leave_inline_fragment(node: PrintedNode, *_args: Any) -> str:
        return join(
            (
                "...",
                wrap("on ", node.type_condition),
                join(node.directives, " "),
                node.selection_set,
            ),
            " ",
        )

    @staticmethod
    def leave_fragment_definition(node: PrintedNode, *_args: Any) -> str:
        # Note: fragment variable definitions are deprecated and will be removed in v3.3
        return (
            f"fragment {node.name}"
            f"{wrap('(', join(node.variable_definitions, ', '), ')')}"
            f" on {node.type_condition}"
            f" {wrap('', join(node.directives, ' '), ' ')}"
            f"{node.selection_set}"
        )

    # Value

    @staticmethod
    def leave_int_value(node: PrintedNode, *_args: Any) -> str:
        return node.value

    @staticmethod
    def leave_float_value(node: PrintedNode, *_args: Any) -> str:
        return node.value

    @staticmethod
    def leave_string_value(node: PrintedNode, *_args: Any) -> str:
        if node.block:
            return print_block_string(node.value)
        return print_string(node.value)

    @staticmethod
    def leave_boolean_value(node: PrintedNode, *_args: Any) -> str:
        return "true" if node.value else "false"

    @staticmethod
    def leave_null_value(_node: PrintedNode, *_args: Any) -> str:
        return "null"

    @staticmethod
    def leave_enum_value(node: PrintedNode, *_args: Any) -> str:
        return node.value

    @staticmethod
    def leave_list_value(node: PrintedNode, *_args: Any) -> str:
        return f"[{join(node.values, ', ')}]"

    @staticmethod
    def leave_object_value(node: PrintedNode, *_args: Any) -> str:
        return f"{{{join(node.fields, ', ')}}}"

    @staticmethod
    def leave_object_field(node: PrintedNode, *_args: Any) -> str:
        return f"{node.name}: {node.value}"

    # Directive

    @staticmethod
    def leave_directive(node: PrintedNode, *_args: Any) -> str:
        return f"@{node.name}{wrap('(', join(node.arguments, ', '), ')')}"

    # Type

    @staticmethod
    def leave_named_type(node: PrintedNode, *_args: Any) -> str:
        return node.name

    @staticmethod
    def leave_list_type(node: PrintedNode, *_args: Any) -> str:
        return f"[{node.type}]"

    @staticmethod
    def leave_non_null_type(node: PrintedNode, *_args: Any) -> str:
        return f"{node.type}!"

    # Type System Definitions

    @staticmethod
    def leave_schema_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            (
                "schema",
                join(node.directives, " "),
                block(node.operation_types),
            ),
            " ",
        )

    @staticmethod
    def leave_operation_type_definition(node: PrintedNode, *_args: Any) -> str:
        return f"{node.operation.value}: {node.type}"

    @staticmethod
    def leave_scalar_type_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            (
                "scalar",
                node.name,
                join(node.directives, " "),
            ),
            " ",
        )

    @staticmethod
    def leave_object_type_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            (
                "type",
                node.name,
                wrap("implements ", join(node.interfaces, " & ")),
                join(node.directives, " "),
                block(node.fields),
            ),
            " ",
        )

    @staticmethod
    def leave_field_definition(node: PrintedNode, *_args: Any) -> str:
        args = node.arguments
        args = (
            wrap("(\n", indent(join(args, "\n")), "\n)")
            if has_multiline_items(args)
            else wrap("(", join(args, ", "), ")")
        )
        directives = wrap(" ", join(node.directives, " "))
        return (
            wrap("", node.description, "\n")
            + f"{node.name}{args}: {node.type}{directives}"
        )

    @staticmethod
    def leave_input_value_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            (
                f"{node.name}: {node.type}",
                wrap("= ", node.default_value),
                join(node.directives, " "),
            ),
            " ",
        )

    @staticmethod
    def leave_interface_type_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            (
                "interface",
                node.name,
                wrap("implements ", join(node.interfaces, " & ")),
                join(node.directives, " "),
                block(node.fields),
            ),
            " ",
        )

    @staticmethod
    def leave_union_type_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            (
                "union",
                node.name,
                join(node.directives, " "),
                wrap("= ", join(node.types, " | ")),
            ),
            " ",
        )

    @staticmethod
    def leave_enum_type_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            ("enum", node.name, join(node.directives, " "), block(node.values)), " "
        )

    @staticmethod
    def leave_enum_value_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            (node.name, join(node.directives, " ")), " "
        )

    @staticmethod
    def leave_input_object_type_definition(node: PrintedNode, *_args: Any) -> str:
        return wrap("", node.description, "\n") + join(
            ("input", node.name, join(node.directives, " "), block(node.fields)), " "
        )

    @staticmethod
    def leave_directive_definition(node: PrintedNode, *_args: Any) -> str:
        args = node.arguments
        args = (
            wrap("(\n", indent(join(args, "\n")), "\n)")
            if has_multiline_items(args)
            else wrap("(", join(args, ", "), ")")
        )
        repeatable = " repeatable" if node.repeatable else ""
        locations = join(node.locations, " | ")
        return (
            wrap("", node.description, "\n")
            + f"directive @{node.name}{args}{repeatable} on {locations}"
        )

    @staticmethod
    def leave_schema_extension(node: PrintedNode, *_args: Any) -> str:
        return join(
            ("extend schema", join(node.directives, " "), block(node.operation_types)),
            " ",
        )

    @staticmethod
    def leave_scalar_type_extension(node: PrintedNode, *_args: Any) -> str:
        return join(("extend scalar", node.name, join(node.directives, " ")), " ")

    @staticmethod
    def leave_object_type_extension(node: PrintedNode, *_args: Any) -> str:
        return join(
            (
                "extend type",
                node.name,
                wrap("implements ", join(node.interfaces, " & ")),
                join(node.directives, " "),
                block(node.fields),
            ),
            " ",
        )

    @staticmethod
    def leave_interface_type_extension(node: PrintedNode, *_args: Any) -> str:
        return join(
            (
                "extend interface",
                node.name,
                wrap("implements ", join(node.interfaces, " & ")),
                join(node.directives, " "),
                block(node.fields),
            ),
            " ",
        )

    @staticmethod
    def leave_union_type_extension(node: PrintedNode, *_args: Any) -> str:
        return join(
            (
                "extend union",
                node.name,
                join(node.directives, " "),
                wrap("= ", join(node.types, " | ")),
            ),
            " ",
        )

    @staticmethod
    def leave_enum_type_extension(node: PrintedNode, *_args: Any) -> str:
        return join(
            ("extend enum", node.name, join(node.directives, " "), block(node.values)),
            " ",
        )

    @staticmethod
    def leave_input_object_type_extension(node: PrintedNode, *_args: Any) -> str:
        return join(
            ("extend input", node.name, join(node.directives, " "), block(node.fields)),
            " ",
        )


def join(strings: Optional[Strings], separator: str = "") -> str:
    """Join strings in a given collection.

    Return an empty string if it is None or empty, otherwise join all items together
    separated by separator if provided.
    """
    return separator.join(s for s in strings if s) if strings else ""


def block(strings: Optional[Strings]) -> str:
    """Return strings inside a block.

    Given a collection of strings, return a string with each item on its own line,
    wrapped in an indented "{ }" block.
    """
    return wrap("{\n", indent(join(strings, "\n")), "\n}")


def wrap(start: str, string: Optional[str], end: str = "") -> str:
    """Wrap string inside other strings at start and end.

    If the string is not None or empty, then wrap with start and end, otherwise return
    an empty string.
    """
    return f"{start}{string}{end}" if string else ""


def indent(string: str) -> str:
    """Indent string with two spaces.

    If the string is not None or empty, add two spaces at the beginning of every line
    inside the string.
    """
    return wrap("  ", string.replace("\n", "\n  "))


def is_multiline(string: str) -> bool:
    """Check whether a string consists of multiple lines."""
    return "\n" in string


def has_multiline_items(strings: Optional[Strings]) -> bool:
    """Check whether one of the items in the list has multiple lines."""
    return any(is_multiline(item) for item in strings) if strings else False



from functools import partial

from graphql.validation import NoUnusedVariablesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, NoUnusedVariablesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_no_unused_variables():
    def uses_all_variables():
        assert_valid(
            """
            query ($a: String, $b: String, $c: String) {
              field(a: $a, b: $b, c: $c)
            }
            """
        )

    def uses_all_variables_deeply():
        assert_valid(
            """
            query Foo($a: String, $b: String, $c: String) {
              field(a: $a) {
                field(b: $b) {
                  field(c: $c)
                }
              }
            }
            """
        )

    def uses_all_variables_deeply_in_inline_fragments():
        assert_valid(
            """
            query Foo($a: String, $b: String, $c: String) {
              ... on Type {
                field(a: $a) {
                  field(b: $b) {
                    ... on Type {
                      field(c: $c)
                    }
                  }
                }
              }
            }
            """
        )

    def uses_all_variables_in_fragment():
        assert_valid(
            """
            query Foo($a: String, $b: String, $c: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a) {
                ...FragB
              }
            }
            fragment FragB on Type {
              field(b: $b) {
                ...FragC
              }
            }
            fragment FragC on Type {
              field(c: $c)
            }
            """
        )

    def variable_used_by_fragment_in_multiple_operations():
        assert_valid(
            """
            query Foo($a: String) {
              ...FragA
            }
            query Bar($b: String) {
              ...FragB
            }
            fragment FragA on Type {
              field(a: $a)
            }
            fragment FragB on Type {
              field(b: $b)
            }
            """
        )

    def variable_used_by_recursive_fragment():
        assert_valid(
            """
            query Foo($a: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a) {
                ...FragA
              }
            }
            """
        )

    def variable_not_used():
        assert_errors(
            """
            query ($a: String, $b: String, $c: String) {
              field(a: $a, b: $b)
            }
            """,
            [{"message": "Variable '$c' is never used.", "locations": [(2, 44)]}],
        )

    def multiple_variables_not_used():
        assert_errors(
            """
            query Foo($a: String, $b: String, $c: String) {
              field(b: $b)
            }
            """,
            [
                {
                    "message": "Variable '$a' is never used in operation 'Foo'.",
                    "locations": [(2, 23)],
                },
                {
                    "message": "Variable '$c' is never used in operation 'Foo'.",
                    "locations": [(2, 47)],
                },
            ],
        )

    def variable_not_used_in_fragments():
        assert_errors(
            """
            query Foo($a: String, $b: String, $c: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a) {
                ...FragB
              }
            }
            fragment FragB on Type {
              field(b: $b) {
                ...FragC
              }
            }
            fragment FragC on Type {
              field
            }
            """,
            [
                {
                    "message": "Variable '$c' is never used in operation 'Foo'.",
                    "locations": [(2, 47)],
                },
            ],
        )

    def multiple_variables_not_used_in_fragments():
        assert_errors(
            """
            query Foo($a: String, $b: String, $c: String) {
              ...FragA
            }
            fragment FragA on Type {
              field {
                ...FragB
              }
            }
            fragment FragB on Type {
              field(b: $b) {
                ...FragC
              }
            }
            fragment FragC on Type {
              field
            }
            """,
            [
                {
                    "message": "Variable '$a' is never used in operation 'Foo'.",
                    "locations": [(2, 23)],
                },
                {
                    "message": "Variable '$c' is never used in operation 'Foo'.",
                    "locations": [(2, 47)],
                },
            ],
        )

    def variable_not_used_by_unreferenced_fragment():
        assert_errors(
            """
            query Foo($b: String) {
              ...FragA
            }
            fragment FragA on Type {
              field(a: $a)
            }
            fragment FragB on Type {
              field(b: $b)
            }
            """,
            [
                {
                    "message": "Variable '$b' is never used in operation 'Foo'.",
                    "locations": [(2, 23)],
                },
            ],
        )

    def variable_not_used_by_fragment_used_by_other_operation():
        assert_errors(
            """
            query Foo($b: String) {
              ...FragA
            }
            query Bar($a: String) {
              ...FragB
            }
            fragment FragA on Type {
              field(a: $a)
            }
            fragment FragB on Type {
              field(b: $b)
            }
            """,
            [
                {
                    "message": "Variable '$b' is never used in operation 'Foo'.",
                    "locations": [(2, 23)],
                },
                {
                    "message": "Variable '$a' is never used in operation 'Bar'.",
                    "locations": [(5, 23)],
                },
            ],
        )



from contextlib import contextmanager
from typing import cast

from pytest import raises

from graphql import graphql_sync
from graphql.type import (
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLObjectType,
    GraphQLNamedType,
    GraphQLSchema,
    GraphQLString,
)
from graphql.pyutils import (
    Description,
    is_description,
    register_description,
    unregister_description,
)
from graphql.utilities import get_introspection_query, print_schema

from ..utils import dedent


class LazyString:
    def __init__(self, text: object) -> None:
        self.text = text

    def __str__(self) -> str:
        return str(self.text)


lazy_string = cast(str, LazyString("Why am I so lazy?"))


@contextmanager
def registered(base: type):
    register_description(base)
    try:
        yield None
    finally:
        unregister_description(LazyString)


def describe_description():
    def by_default_strings_are_accepted():
        is_description("")
        is_description("text")

    def by_default_non_strings_are_not_accepted():
        assert not is_description(None)
        assert not is_description(b"bytes")
        assert not is_description(0)
        assert not is_description(42)
        assert not is_description(("tuple",))
        assert not is_description(["list"])

    def after_registration_lazy_strings_are_accepted():
        with registered(LazyString):
            assert is_description("not lazy")
            assert is_description(lazy_string)
            assert not is_description(42)

    def can_register_and_unregister():
        try:
            assert Description.bases is str
            register_description(str)
            assert Description.bases is str
            register_description(int)
            assert Description.bases == (str, int)
            register_description(int)
            assert Description.bases == (str, int)
            register_description(float)
            assert Description.bases == (str, int, float)
            unregister_description(int)
            assert Description.bases == (str, float)
            unregister_description(float)
            assert Description.bases is str
            unregister_description(str)
            assert Description.bases is object
            register_description(str)
            assert Description.bases is str
            register_description(object)
            assert Description.bases is object
            Description.bases = (str,)
            unregister_description(str)
            assert Description.bases is object
            unregister_description(str)
            assert Description.bases is object
        finally:
            Description.bases = str

    def can_only_register_types():
        with raises(TypeError, match="Only types can be registered\\."):
            # noinspection PyTypeChecker
            register_description("foo")  # type: ignore

    def can_only_unregister_types():
        with raises(TypeError, match="Only types can be unregistered\\."):
            # noinspection PyTypeChecker
            unregister_description("foo")  # type: ignore

    def describe_graphql_types():
        def graphql_named_type():
            named_type = GraphQLNamedType(name="Foo", description="not lazy")
            assert named_type.name == "Foo"
            assert named_type.description == "not lazy"
            with raises(TypeError, match="Expected name to be a string\\."):
                GraphQLNamedType(name=lazy_string)
            with raises(TypeError, match="The description must be a string\\."):
                GraphQLNamedType(name="Foo", description=lazy_string)
            with registered(LazyString):
                named_type = GraphQLNamedType(name="Foo", description=lazy_string)
                assert named_type.description is lazy_string
                assert str(named_type.description).endswith("lazy?")
                with raises(TypeError, match="Expected name to be a string\\."):
                    GraphQLNamedType(name=lazy_string)

        def graphql_field():
            field = GraphQLField(GraphQLString, description="not lazy")
            assert field.description == "not lazy"
            field = GraphQLField(GraphQLString, deprecation_reason="not lazy")
            assert field.deprecation_reason == "not lazy"
            with raises(TypeError, match="The description must be a string\\."):
                GraphQLField(GraphQLString, description=lazy_string)
            with raises(TypeError, match="The deprecation reason must be a string\\."):
                GraphQLField(GraphQLString, deprecation_reason=lazy_string)
            with registered(LazyString):
                field = GraphQLField(
                    GraphQLString,
                    description=lazy_string,
                    deprecation_reason=lazy_string,
                )
                assert field.description is lazy_string
                assert str(field.description).endswith("lazy?")
                assert field.deprecation_reason is lazy_string
                assert str(field.deprecation_reason).endswith("lazy?")

        def graphql_argument():
            arg = GraphQLArgument(GraphQLString, description="not lazy")
            assert arg.description == "not lazy"
            with raises(TypeError, match="Argument description must be a string\\."):
                GraphQLArgument(GraphQLString, description=lazy_string)
            with registered(LazyString):
                arg = GraphQLArgument(GraphQLString, description=lazy_string)
                assert arg.description is lazy_string
                assert str(arg.description).endswith("lazy?")

        def graphql_enum_value():
            value = GraphQLEnumValue(description="not lazy")
            assert value.description == "not lazy"
            value = GraphQLEnumValue(deprecation_reason="not lazy")
            assert value.deprecation_reason == "not lazy"
            with raises(
                TypeError, match="The description of the enum value must be a string\\."
            ):
                GraphQLEnumValue(description=lazy_string)
            with raises(
                TypeError,
                match="The deprecation reason for the enum value must be a string\\.",
            ):
                GraphQLEnumValue(deprecation_reason=lazy_string)
            with registered(LazyString):
                value = GraphQLEnumValue(
                    description=lazy_string, deprecation_reason=lazy_string
                )
                assert value.description is lazy_string
                assert str(value.description).endswith("lazy?")
                assert value.deprecation_reason is lazy_string
                assert str(value.deprecation_reason).endswith("lazy?")

        def graphql_input_field():
            field = GraphQLInputField(GraphQLString, description="not lazy")
            assert field.description == "not lazy"
            with raises(TypeError, match="Input field description must be a string\\."):
                GraphQLInputField(GraphQLString, description=lazy_string)
            with registered(LazyString):
                field = GraphQLInputField(GraphQLString, description=lazy_string)
                assert field.description is lazy_string
                assert str(field.description).endswith("lazy?")

        def graphql_directive():
            directive = GraphQLDirective("Foo", [], description="not lazy")
            assert directive.name == "Foo"
            assert directive.description == "not lazy"
            with raises(TypeError, match="Expected name to be a string\\."):
                GraphQLDirective(lazy_string, [])
            with raises(TypeError, match="Foo description must be a string\\."):
                GraphQLDirective("Foo", [], description=lazy_string)
            with registered(LazyString):
                directive = GraphQLDirective("Foo", [], description=lazy_string)
                assert directive.description is lazy_string
                assert str(directive.description).endswith("lazy?")
                with raises(TypeError, match="Expected name to be a string\\."):
                    GraphQLDirective(lazy_string, [])

    def handels_introspection():
        class Lazy:
            def __init__(self, text: str):
                self.text = text
                self.evaluated = False

            def __str__(self) -> str:
                self.evaluated = True
                return self.text

        description = Lazy("a lazy description")
        deprecation_reason = Lazy("a lazy reason")

        with registered(Lazy):
            field = GraphQLField(
                GraphQLString,
                description=cast(str, description),
                deprecation_reason=cast(str, deprecation_reason),
            )

        schema = GraphQLSchema(GraphQLObjectType("Query", {"lazyField": field}))

        query = get_introspection_query(descriptions=True)
        assert not description.evaluated
        assert not deprecation_reason.evaluated
        result = graphql_sync(schema, query)
        assert description.evaluated
        assert deprecation_reason.evaluated
        assert result.data
        introspected_query = result.data["__schema"]["types"][0]
        assert introspected_query["name"] == "Query"
        introspected_field = introspected_query["fields"][0]
        assert introspected_field["name"] == "lazyField"
        assert introspected_field["description"] == "a lazy description"
        assert introspected_field["deprecationReason"] == "a lazy reason"

    def handles_printing():
        class Lazy:
            def __init__(self, text: str):
                self.text = text
                self.evaluated = False

            def __str__(self) -> str:
                self.evaluated = True
                return self.text

        description = Lazy("a lazy description")
        deprecation_reason = Lazy("a lazy reason")

        with registered(Lazy):
            field = GraphQLField(
                GraphQLString,
                description=cast(str, description),
                deprecation_reason=cast(str, deprecation_reason),
            )

        schema = GraphQLSchema(GraphQLObjectType("Query", {"lazyField": field}))

        assert not description.evaluated
        assert not deprecation_reason.evaluated
        assert print_schema(schema) == dedent(
            '''
            type Query {
              """a lazy description"""
              lazyField: String @deprecated(reason: "a lazy reason")
            }
            '''
        )
        assert description.evaluated
        assert deprecation_reason.evaluated



from typing import Any, Callable, Dict, List, Optional, Union, cast

from ..language import print_ast, StringValueNode
from ..language.block_string import is_printable_as_block_string
from ..pyutils import inspect
from ..type import (
    DEFAULT_DEPRECATION_REASON,
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInterfaceType,
    GraphQLNamedType,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLUnionType,
    is_enum_type,
    is_input_object_type,
    is_interface_type,
    is_introspection_type,
    is_object_type,
    is_scalar_type,
    is_specified_directive,
    is_specified_scalar_type,
    is_union_type,
)
from .ast_from_value import ast_from_value

__all__ = ["print_schema", "print_introspection_schema", "print_type", "print_value"]


def print_schema(schema: GraphQLSchema) -> str:
    return print_filtered_schema(
        schema, lambda n: not is_specified_directive(n), is_defined_type
    )


def print_introspection_schema(schema: GraphQLSchema) -> str:
    return print_filtered_schema(schema, is_specified_directive, is_introspection_type)


def is_defined_type(type_: GraphQLNamedType) -> bool:
    return not is_specified_scalar_type(type_) and not is_introspection_type(type_)


def print_filtered_schema(
    schema: GraphQLSchema,
    directive_filter: Callable[[GraphQLDirective], bool],
    type_filter: Callable[[GraphQLNamedType], bool],
) -> str:
    directives = filter(directive_filter, schema.directives)
    types = filter(type_filter, schema.type_map.values())

    return "\n\n".join(
        (
            *filter(None, (print_schema_definition(schema),)),
            *map(print_directive, directives),
            *map(print_type, types),
        )
    )


def print_schema_definition(schema: GraphQLSchema) -> Optional[str]:
    if schema.description is None and is_schema_of_common_names(schema):
        return None

    operation_types = []

    query_type = schema.query_type
    if query_type:
        operation_types.append(f"  query: {query_type.name}")

    mutation_type = schema.mutation_type
    if mutation_type:
        operation_types.append(f"  mutation: {mutation_type.name}")

    subscription_type = schema.subscription_type
    if subscription_type:
        operation_types.append(f"  subscription: {subscription_type.name}")

    return print_description(schema) + "schema {\n" + "\n".join(operation_types) + "\n}"


def is_schema_of_common_names(schema: GraphQLSchema) -> bool:
    """Check whether this schema uses the common naming convention.

    GraphQL schema define root types for each type of operation. These types are the
    same as any other type and can be named in any manner, however there is a common
    naming convention:

    schema {
      query: Query
      mutation: Mutation
      subscription: Subscription
    }

    When using this naming convention, the schema description can be omitted.
    """
    query_type = schema.query_type
    if query_type and query_type.name != "Query":
        return False

    mutation_type = schema.mutation_type
    if mutation_type and mutation_type.name != "Mutation":
        return False

    subscription_type = schema.subscription_type
    return not subscription_type or subscription_type.name == "Subscription"


def print_type(type_: GraphQLNamedType) -> str:
    if is_scalar_type(type_):
        type_ = cast(GraphQLScalarType, type_)
        return print_scalar(type_)
    if is_object_type(type_):
        type_ = cast(GraphQLObjectType, type_)
        return print_object(type_)
    if is_interface_type(type_):
        type_ = cast(GraphQLInterfaceType, type_)
        return print_interface(type_)
    if is_union_type(type_):
        type_ = cast(GraphQLUnionType, type_)
        return print_union(type_)
    if is_enum_type(type_):
        type_ = cast(GraphQLEnumType, type_)
        return print_enum(type_)
    if is_input_object_type(type_):
        type_ = cast(GraphQLInputObjectType, type_)
        return print_input_object(type_)

    # Not reachable. All possible types have been considered.
    raise TypeError(f"Unexpected type: {inspect(type_)}.")


def print_scalar(type_: GraphQLScalarType) -> str:
    return (
        print_description(type_)
        + f"scalar {type_.name}"
        + print_specified_by_url(type_)
    )


def print_implemented_interfaces(
    type_: Union[GraphQLObjectType, GraphQLInterfaceType]
) -> str:
    interfaces = type_.interfaces
    return " implements " + " & ".join(i.name for i in interfaces) if interfaces else ""


def print_object(type_: GraphQLObjectType) -> str:
    return (
        print_description(type_)
        + f"type {type_.name}"
        + print_implemented_interfaces(type_)
        + print_fields(type_)
    )


def print_interface(type_: GraphQLInterfaceType) -> str:
    return (
        print_description(type_)
        + f"interface {type_.name}"
        + print_implemented_interfaces(type_)
        + print_fields(type_)
    )


def print_union(type_: GraphQLUnionType) -> str:
    types = type_.types
    possible_types = " = " + " | ".join(t.name for t in types) if types else ""
    return print_description(type_) + f"union {type_.name}" + possible_types


def print_enum(type_: GraphQLEnumType) -> str:
    values = [
        print_description(value, "  ", not i)
        + f"  {name}"
        + print_deprecated(value.deprecation_reason)
        for i, (name, value) in enumerate(type_.values.items())
    ]
    return print_description(type_) + f"enum {type_.name}" + print_block(values)


def print_input_object(type_: GraphQLInputObjectType) -> str:
    fields = [
        print_description(field, "  ", not i) + "  " + print_input_value(name, field)
        for i, (name, field) in enumerate(type_.fields.items())
    ]
    return print_description(type_) + f"input {type_.name}" + print_block(fields)


def print_fields(type_: Union[GraphQLObjectType, GraphQLInterfaceType]) -> str:
    fields = [
        print_description(field, "  ", not i)
        + f"  {name}"
        + print_args(field.args, "  ")
        + f": {field.type}"
        + print_deprecated(field.deprecation_reason)
        for i, (name, field) in enumerate(type_.fields.items())
    ]
    return print_block(fields)


def print_block(items: List[str]) -> str:
    return " {\n" + "\n".join(items) + "\n}" if items else ""


def print_args(args: Dict[str, GraphQLArgument], indentation: str = "") -> str:
    if not args:
        return ""

    # If every arg does not have a description, print them on one line.
    if not any(arg.description for arg in args.values()):
        return (
            "("
            + ", ".join(print_input_value(name, arg) for name, arg in args.items())
            + ")"
        )

    return (
        "(\n"
        + "\n".join(
            print_description(arg, f"  {indentation}", not i)
            + f"  {indentation}"
            + print_input_value(name, arg)
            for i, (name, arg) in enumerate(args.items())
        )
        + f"\n{indentation})"
    )


def print_input_value(name: str, arg: GraphQLArgument) -> str:
    default_ast = ast_from_value(arg.default_value, arg.type)
    arg_decl = f"{name}: {arg.type}"
    if default_ast:
        arg_decl += f" = {print_ast(default_ast)}"
    return arg_decl + print_deprecated(arg.deprecation_reason)


def print_directive(directive: GraphQLDirective) -> str:
    return (
        print_description(directive)
        + f"directive @{directive.name}"
        + print_args(directive.args)
        + (" repeatable" if directive.is_repeatable else "")
        + " on "
        + " | ".join(location.name for location in directive.locations)
    )


def print_deprecated(reason: Optional[str]) -> str:
    if reason is None:
        return ""
    if reason != DEFAULT_DEPRECATION_REASON:
        ast_value = print_ast(StringValueNode(value=reason))
        return f" @deprecated(reason: {ast_value})"
    return " @deprecated"


def print_specified_by_url(scalar: GraphQLScalarType) -> str:
    if scalar.specified_by_url is None:
        return ""
    ast_value = print_ast(StringValueNode(value=scalar.specified_by_url))
    return f" @specifiedBy(url: {ast_value})"


def print_description(
    def_: Union[
        GraphQLArgument,
        GraphQLDirective,
        GraphQLEnumValue,
        GraphQLNamedType,
        GraphQLSchema,
    ],
    indentation: str = "",
    first_in_block: bool = True,
) -> str:
    description = def_.description
    if description is None:
        return ""

    block_string = print_ast(
        StringValueNode(
            value=description, block=is_printable_as_block_string(description)
        )
    )

    prefix = "\n" + indentation if indentation and not first_in_block else indentation

    return prefix + block_string.replace("\n", "\n" + indentation) + "\n"


def print_value(value: Any, type_: GraphQLInputType) -> str:
    """@deprecated: Convenience function for printing a Python value"""
    return print_ast(ast_from_value(value, type_))  # type: ignore



from itertools import product
from typing import Generator

__all__ = ["gen_fuzz_strings"]


def gen_fuzz_strings(allowed_chars: str, max_length: int) -> Generator[str, None, None]:
    """Generator that produces all possible combinations of allowed characters."""
    for length in range(max_length + 1):
        yield from map("".join, product(allowed_chars, repeat=length))



from typing import List, NamedTuple, Optional

from ..error import GraphQLSyntaxError
from .ast import Token
from .block_string import dedent_block_string_lines
from .character_classes import is_digit, is_name_start, is_name_continue
from .source import Source
from .token_kind import TokenKind

__all__ = ["Lexer", "is_punctuator_token_kind"]


class EscapeSequence(NamedTuple):
    """The string value and lexed size of an escape sequence."""

    value: str
    size: int


class Lexer:
    """GraphQL Lexer

    A Lexer is a stateful stream generator in that every time it is advanced, it returns
    the next token in the Source. Assuming the source lexes, the final Token emitted by
    the lexer will be of kind EOF, after which the lexer will repeatedly return the same
    EOF token whenever called.
    """

    def __init__(self, source: Source):
        """Given a Source object, initialize a Lexer for that source."""
        self.source = source
        self.token = self.last_token = Token(TokenKind.SOF, 0, 0, 0, 0)
        self.line, self.line_start = 1, 0

    def advance(self) -> Token:
        """Advance the token stream to the next non-ignored token."""
        self.last_token = self.token
        token = self.token = self.lookahead()
        return token

    def lookahead(self) -> Token:
        """Look ahead and return the next non-ignored token, but do not change state."""
        token = self.token
        if token.kind != TokenKind.EOF:
            while True:
                if token.next:
                    token = token.next
                else:
                    # Read the next token and form a link in the token linked-list.
                    next_token = self.read_next_token(token.end)
                    token.next = next_token
                    next_token.prev = token
                    token = next_token
                if token.kind != TokenKind.COMMENT:
                    break
        return token

    def print_code_point_at(self, location: int) -> str:
        """Print the code point at the given location.

        Prints the code point (or end of file reference) at a given location in a
        source for use in error messages.

        Printable ASCII is printed quoted, while other points are printed in Unicode
        code point form (ie. U+1234).
        """
        body = self.source.body
        if location >= len(body):
            return TokenKind.EOF.value
        char = body[location]
        # Printable ASCII
        if "\x20" <= char <= "\x7E":
            return "'\"'" if char == '"' else f"'{char}'"
        # Unicode code point
        point = ord(
            body[location : location + 2]
            .encode("utf-16", "surrogatepass")
            .decode("utf-16")
            if is_supplementary_code_point(body, location)
            else char
        )
        return f"U+{point:04X}"

    def create_token(
        self, kind: TokenKind, start: int, end: int, value: Optional[str] = None
    ) -> Token:
        """Create a token with line and column location information."""
        line = self.line
        col = 1 + start - self.line_start
        return Token(kind, start, end, line, col, value)

    def read_next_token(self, start: int) -> Token:
        """Get the next token from the source starting at the given position.

        This skips over whitespace until it finds the next lexable token, then lexes
        punctuators immediately or calls the appropriate helper function for more
        complicated tokens.
        """
        body = self.source.body
        body_length = len(body)
        position = start

        while position < body_length:
            char = body[position]  # SourceCharacter

            if char in " \t,\ufeff":
                position += 1
                continue
            elif char == "\n":
                position += 1
                self.line += 1
                self.line_start = position
                continue
            elif char == "\r":
                if body[position + 1 : position + 2] == "\n":
                    position += 2
                else:
                    position += 1
                self.line += 1
                self.line_start = position
                continue

            if char == "#":
                return self.read_comment(position)

            if char == '"':
                if body[position + 1 : position + 3] == '""':
                    return self.read_block_string(position)
                return self.read_string(position)

            kind = _KIND_FOR_PUNCT.get(char)
            if kind:
                return self.create_token(kind, position, position + 1)

            if is_digit(char) or char == "-":
                return self.read_number(position, char)

            if is_name_start(char):
                return self.read_name(position)

            if char == ".":
                if body[position + 1 : position + 3] == "..":
                    return self.create_token(TokenKind.SPREAD, position, position + 3)

            message = (
                "Unexpected single quote character ('),"
                ' did you mean to use a double quote (")?'
                if char == "'"
                else (
                    f"Unexpected character: {self.print_code_point_at(position)}."
                    if is_unicode_scalar_value(char)
                    or is_supplementary_code_point(body, position)
                    else f"Invalid character: {self.print_code_point_at(position)}."
                )
            )

            raise GraphQLSyntaxError(self.source, position, message)

        return self.create_token(TokenKind.EOF, body_length, body_length)

    def read_comment(self, start: int) -> Token:
        """Read a comment token from the source file."""
        body = self.source.body
        body_length = len(body)

        position = start + 1
        while position < body_length:
            char = body[position]
            if char in "\r\n":
                break
            if is_unicode_scalar_value(char):
                position += 1
            elif is_supplementary_code_point(body, position):
                position += 2
            else:
                break  # pragma: no cover

        return self.create_token(
            TokenKind.COMMENT,
            start,
            position,
            body[start + 1 : position],
        )

    def read_number(self, start: int, first_char: str) -> Token:
        """Reads a number token from the source file.

        This can be either a FloatValue or an IntValue,
        depending on whether a FractionalPart or ExponentPart is encountered.
        """
        body = self.source.body
        position = start
        char = first_char
        is_float = False

        if char == "-":
            position += 1
            char = body[position : position + 1]
        if char == "0":
            position += 1
            char = body[position : position + 1]
            if is_digit(char):
                raise GraphQLSyntaxError(
                    self.source,
                    position,
                    "Invalid number, unexpected digit after 0:"
                    f" {self.print_code_point_at(position)}.",
                )
        else:
            position = self.read_digits(position, char)
            char = body[position : position + 1]
        if char == ".":
            is_float = True
            position += 1
            char = body[position : position + 1]
            position = self.read_digits(position, char)
            char = body[position : position + 1]
        if char and char in "Ee":
            is_float = True
            position += 1
            char = body[position : position + 1]
            if char and char in "+-":
                position += 1
                char = body[position : position + 1]
            position = self.read_digits(position, char)
            char = body[position : position + 1]

        # Numbers cannot be followed by . or NameStart
        if char and (char == "." or is_name_start(char)):
            raise GraphQLSyntaxError(
                self.source,
                position,
                "Invalid number, expected digit but got:"
                f" {self.print_code_point_at(position)}.",
            )

        return self.create_token(
            TokenKind.FLOAT if is_float else TokenKind.INT,
            start,
            position,
            body[start:position],
        )

    def read_digits(self, start: int, first_char: str) -> int:
        """Return the new position in the source after reading one or more digits."""
        if not is_digit(first_char):
            raise GraphQLSyntaxError(
                self.source,
                start,
                "Invalid number, expected digit but got:"
                f" {self.print_code_point_at(start)}.",
            )

        body = self.source.body
        body_length = len(body)
        position = start + 1
        while position < body_length and is_digit(body[position]):
            position += 1
        return position

    def read_string(self, start: int) -> Token:
        """Read a single-quote string token from the source file."""
        body = self.source.body
        body_length = len(body)
        position = start + 1
        chunk_start = position
        value: List[str] = []
        append = value.append

        while position < body_length:
            char = body[position]

            if char == '"':
                append(body[chunk_start:position])
                return self.create_token(
                    TokenKind.STRING,
                    start,
                    position + 1,
                    "".join(value),
                )

            if char == "\\":
                append(body[chunk_start:position])
                escape = (
                    (
                        self.read_escaped_unicode_variable_width(position)
                        if body[position + 2 : position + 3] == "{"
                        else self.read_escaped_unicode_fixed_width(position)
                    )
                    if body[position + 1 : position + 2] == "u"
                    else self.read_escaped_character(position)
                )
                append(escape.value)
                position += escape.size
                chunk_start = position
                continue

            if char in "\r\n":
                break

            if is_unicode_scalar_value(char):
                position += 1
            elif is_supplementary_code_point(body, position):
                position += 2
            else:
                raise GraphQLSyntaxError(
                    self.source,
                    position,
                    "Invalid character within String:"
                    f" {self.print_code_point_at(position)}.",
                )

        raise GraphQLSyntaxError(self.source, position, "Unterminated string.")

    def read_escaped_unicode_variable_width(self, position: int) -> EscapeSequence:
        body = self.source.body
        point = 0
        size = 3
        max_size = min(12, len(body) - position)
        # Cannot be larger than 12 chars (\u{00000000}).
        while size < max_size:
            char = body[position + size]
            size += 1
            if char == "}":
                # Must be at least 5 chars (\u{0}) and encode a Unicode scalar value.
                if size < 5 or not (
                    0 <= point <= 0xD7FF or 0xE000 <= point <= 0x10FFFF
                ):
                    break
                return EscapeSequence(chr(point), size)
            # Append this hex digit to the code point.
            point = (point << 4) | read_hex_digit(char)
            if point < 0:
                break

        raise GraphQLSyntaxError(
            self.source,
            position,
            f"Invalid Unicode escape sequence: '{body[position: position + size]}'.",
        )

    def read_escaped_unicode_fixed_width(self, position: int) -> EscapeSequence:
        body = self.source.body
        code = read_16_bit_hex_code(body, position + 2)

        if 0 <= code <= 0xD7FF or 0xE000 <= code <= 0x10FFFF:
            return EscapeSequence(chr(code), 6)

        # GraphQL allows JSON-style surrogate pair escape sequences, but only when
        # a valid pair is formed.
        if 0xD800 <= code <= 0xDBFF:
            if body[position + 6 : position + 8] == "\\u":
                trailing_code = read_16_bit_hex_code(body, position + 8)
                if 0xDC00 <= trailing_code <= 0xDFFF:
                    return EscapeSequence(
                        (chr(code) + chr(trailing_code))
                        .encode("utf-16", "surrogatepass")
                        .decode("utf-16"),
                        12,
                    )

        raise GraphQLSyntaxError(
            self.source,
            position,
            f"Invalid Unicode escape sequence: '{body[position: position + 6]}'.",
        )

    def read_escaped_character(self, position: int) -> EscapeSequence:
        body = self.source.body
        value = _ESCAPED_CHARS.get(body[position + 1])
        if value:
            return EscapeSequence(value, 2)
        raise GraphQLSyntaxError(
            self.source,
            position,
            f"Invalid character escape sequence: '{body[position: position + 2]}'.",
        )

    def read_block_string(self, start: int) -> Token:
        """Read a block string token from the source file."""
        body = self.source.body
        body_length = len(body)
        line_start = self.line_start

        position = start + 3
        chunk_start = position
        current_line = ""

        block_lines = []
        while position < body_length:
            char = body[position]

            if char == '"' and body[position + 1 : position + 3] == '""':
                current_line += body[chunk_start:position]
                block_lines.append(current_line)

                token = self.create_token(
                    TokenKind.BLOCK_STRING,
                    start,
                    position + 3,
                    # return a string of the lines joined with new lines
                    "\n".join(dedent_block_string_lines(block_lines)),
                )

                self.line += len(block_lines) - 1
                self.line_start = line_start
                return token

            if char == "\\" and body[position + 1 : position + 4] == '"""':
                current_line += body[chunk_start:position]
                chunk_start = position + 1  # skip only slash
                position += 4
                continue

            if char in "\r\n":
                current_line += body[chunk_start:position]
                block_lines.append(current_line)

                if char == "\r" and body[position + 1 : position + 2] == "\n":
                    position += 2
                else:
                    position += 1

                current_line = ""
                chunk_start = line_start = position
                continue

            if is_unicode_scalar_value(char):
                position += 1
            elif is_supplementary_code_point(body, position):
                position += 2
            else:
                raise GraphQLSyntaxError(
                    self.source,
                    position,
                    "Invalid character within String:"
                    f" {self.print_code_point_at(position)}.",
                )

        raise GraphQLSyntaxError(self.source, position, "Unterminated string.")

    def read_name(self, start: int) -> Token:
        """Read an alphanumeric + underscore name from the source."""
        body = self.source.body
        body_length = len(body)
        position = start + 1

        while position < body_length:
            char = body[position]
            if not is_name_continue(char):
                break
            position += 1

        return self.create_token(TokenKind.NAME, start, position, body[start:position])


_punctuator_token_kinds = frozenset(
    [
        TokenKind.BANG,
        TokenKind.DOLLAR,
        TokenKind.AMP,
        TokenKind.PAREN_L,
        TokenKind.PAREN_R,
        TokenKind.SPREAD,
        TokenKind.COLON,
        TokenKind.EQUALS,
        TokenKind.AT,
        TokenKind.BRACKET_L,
        TokenKind.BRACKET_R,
        TokenKind.BRACE_L,
        TokenKind.PIPE,
        TokenKind.BRACE_R,
    ]
)


def is_punctuator_token_kind(kind: TokenKind) -> bool:
    """Check whether the given token kind corresponds to a punctuator.

    For internal use only.
    """
    return kind in _punctuator_token_kinds


_KIND_FOR_PUNCT = {
    "!": TokenKind.BANG,
    "$": TokenKind.DOLLAR,
    "&": TokenKind.AMP,
    "(": TokenKind.PAREN_L,
    ")": TokenKind.PAREN_R,
    ":": TokenKind.COLON,
    "=": TokenKind.EQUALS,
    "@": TokenKind.AT,
    "[": TokenKind.BRACKET_L,
    "]": TokenKind.BRACKET_R,
    "{": TokenKind.BRACE_L,
    "}": TokenKind.BRACE_R,
    "|": TokenKind.PIPE,
}


_ESCAPED_CHARS = {
    '"': '"',
    "/": "/",
    "\\": "\\",
    "b": "\b",
    "f": "\f",
    "n": "\n",
    "r": "\r",
    "t": "\t",
}


def read_16_bit_hex_code(body: str, position: int) -> int:
    """Read a 16bit hexadecimal string and return its positive integer value (0-65535).

    Reads four hexadecimal characters and returns the positive integer that 16bit
    hexadecimal string represents. For example, "000f" will return 15, and "dead"
    will return 57005.

    Returns a negative number if any char was not a valid hexadecimal digit.
    """
    # read_hex_digit() returns -1 on error. ORing a negative value with any other
    # value always produces a negative value.
    return (
        read_hex_digit(body[position]) << 12
        | read_hex_digit(body[position + 1]) << 8
        | read_hex_digit(body[position + 2]) << 4
        | read_hex_digit(body[position + 3])
    )


def read_hex_digit(char: str) -> int:
    """Read a hexadecimal character and returns its positive integer value (0-15).

    '0' becomes 0, '9' becomes 9
    'A' becomes 10, 'F' becomes 15
    'a' becomes 10, 'f' becomes 15

    Returns -1 if the provided character code was not a valid hexadecimal digit.
    """
    if "0" <= char <= "9":
        return ord(char) - 48
    elif "A" <= char <= "F":
        return ord(char) - 55
    elif "a" <= char <= "f":
        return ord(char) - 87
    return -1


def is_unicode_scalar_value(char: str) -> bool:
    """Check whether this is a Unicode scalar value.

    A Unicode scalar value is any Unicode code point except surrogate code
    points. In other words, the inclusive ranges of values 0x0000 to 0xD7FF and
    0xE000 to 0x10FFFF.
    """
    return "\x00" <= char <= "\ud7ff" or "\ue000" <= char <= "\U0010ffff"


def is_supplementary_code_point(body: str, location: int) -> bool:
    """
    Check whether the current location is a supplementary code point.

    The GraphQL specification defines source text as a sequence of unicode scalar
    values (which Unicode defines to exclude surrogate code points).
    """
    try:
        return (
            "\ud800" <= body[location] <= "\udbff"
            and "\udc00" <= body[location + 1] <= "\udfff"
        )
    except IndexError:
        return False



from functools import partial

from graphql.utilities import build_schema
from graphql.validation.rules.unique_directive_names import UniqueDirectiveNamesRule

from .harness import assert_sdl_validation_errors

assert_errors = partial(assert_sdl_validation_errors, UniqueDirectiveNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_directive_names():
    def no_directive():
        assert_valid(
            """
            type Foo
            """
        )

    def one_directive():
        assert_valid(
            """
            directive @foo on SCHEMA
            """
        )

    def many_directives():
        assert_valid(
            """
            directive @foo on SCHEMA
            directive @bar on SCHEMA
            directive @baz on SCHEMA
            """
        )

    def directive_and_non_directive_definitions_named_the_same():
        assert_valid(
            """
            query foo { __typename }
            fragment foo on foo { __typename }
            type foo

            directive @foo on SCHEMA
            """
        )

    def directives_named_the_same():
        assert_errors(
            """
            directive @foo on SCHEMA

            directive @foo on SCHEMA
            """,
            [
                {
                    "message": "There can be only one directive named '@foo'.",
                    "locations": [(2, 24), (4, 24)],
                }
            ],
        )

    def adding_new_directive_to_existing_schema():
        schema = build_schema("directive @foo on SCHEMA")

        assert_valid("directive @bar on SCHEMA", schema=schema)

    def adding_new_directive_with_standard_name_to_existing_schema():
        schema = build_schema("type foo")

        assert_errors(
            "directive @skip on SCHEMA",
            [
                {
                    "message": "Directive '@skip' already exists in the schema."
                    " It cannot be redefined.",
                    "locations": [(1, 12)],
                }
            ],
            schema,
        )

    def adding_new_directive_to_existing_schema_with_same_named_type():
        schema = build_schema("type foo")

        assert_valid("directive @foo on SCHEMA", schema=schema)

    def adding_conflicting_directives_to_existing_schema():
        schema = build_schema("directive @foo on SCHEMA")

        assert_errors(
            "directive @foo on SCHEMA",
            [
                {
                    "message": "Directive '@foo' already exists in the schema."
                    " It cannot be redefined.",
                    "locations": [(1, 12)],
                }
            ],
            schema,
        )



import asyncio
from typing import Awaitable

from pytest import mark

from graphql.execution import execute
from graphql.language import parse
from graphql.type import (
    GraphQLSchema,
    GraphQLObjectType,
    GraphQLField,
    GraphQLList,
    GraphQLInterfaceType,
    GraphQLBoolean,
    GraphQLInt,
    GraphQLString,
)


class Barrier:
    """Barrier that makes progress only after a certain number of waits."""

    def __init__(self, number: int):
        self.event = asyncio.Event()
        self.number = number

    async def wait(self) -> bool:
        self.number -= 1
        if not self.number:
            self.event.set()
        return await self.event.wait()


def describe_parallel_execution():
    @mark.asyncio
    async def resolve_fields_in_parallel():
        barrier = Barrier(2)

        async def resolve(*_args):
            return await barrier.wait()

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "foo": GraphQLField(GraphQLBoolean, resolve=resolve),
                    "bar": GraphQLField(GraphQLBoolean, resolve=resolve),
                },
            )
        )

        ast = parse("{foo, bar}")

        # raises TimeoutError if not parallel
        awaitable_result = execute(schema, ast)
        assert isinstance(awaitable_result, Awaitable)
        result = await asyncio.wait_for(awaitable_result, 1.0)

        assert result == ({"foo": True, "bar": True}, None)

    @mark.asyncio
    async def resolve_list_in_parallel():
        barrier = Barrier(2)

        async def resolve(*_args):
            return await barrier.wait()

        async def resolve_list(*args):
            return [resolve(*args), resolve(*args)]

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "foo": GraphQLField(
                        GraphQLList(GraphQLBoolean), resolve=resolve_list
                    )
                },
            )
        )

        ast = parse("{foo}")

        # raises TimeoutError if not parallel
        awaitable_result = execute(schema, ast)
        assert isinstance(awaitable_result, Awaitable)
        result = await asyncio.wait_for(awaitable_result, 1.0)

        assert result == ({"foo": [True, True]}, None)

    @mark.asyncio
    async def resolve_is_type_of_in_parallel():
        FooType = GraphQLInterfaceType("Foo", {"foo": GraphQLField(GraphQLString)})

        barrier = Barrier(4)

        async def is_type_of_bar(obj, *_args):
            await barrier.wait()
            return obj["foo"] == "bar"

        BarType = GraphQLObjectType(
            "Bar",
            {"foo": GraphQLField(GraphQLString), "foobar": GraphQLField(GraphQLInt)},
            interfaces=[FooType],
            is_type_of=is_type_of_bar,
        )

        async def is_type_of_baz(obj, *_args):
            await barrier.wait()
            return obj["foo"] == "baz"

        BazType = GraphQLObjectType(
            "Baz",
            {"foo": GraphQLField(GraphQLString), "foobaz": GraphQLField(GraphQLInt)},
            interfaces=[FooType],
            is_type_of=is_type_of_baz,
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "foo": GraphQLField(
                        GraphQLList(FooType),
                        resolve=lambda *_args: [
                            {"foo": "bar", "foobar": 1},
                            {"foo": "baz", "foobaz": 2},
                        ],
                    )
                },
            ),
            types=[BarType, BazType],
        )

        ast = parse(
            """
            {
              foo {
                foo
                ... on Bar { foobar }
                ... on Baz { foobaz }
              }
            }
            """
        )

        # raises TimeoutError if not parallel
        awaitable_result = execute(schema, ast)
        assert isinstance(awaitable_result, Awaitable)
        result = await asyncio.wait_for(awaitable_result, 1.0)

        assert result == (
            {"foo": [{"foo": "bar", "foobar": 1}, {"foo": "baz", "foobaz": 2}]},
            None,
        )



from typing import cast, Any, Dict

from graphql.language import DirectiveLocation
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLField,
    GraphQLFloat,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLString,
    GraphQLUnionType,
    GraphQLInputField,
    GraphQLDirective,
)
from graphql.utilities import (
    build_schema,
    print_schema,
    print_introspection_schema,
    print_value,
)

from ..utils import dedent


def expect_printed_schema(schema: GraphQLSchema) -> str:
    schema_text = print_schema(schema)
    # keep print_schema and build_schema in sync
    assert print_schema(build_schema(schema_text)) == schema_text
    return schema_text


def build_single_field_schema(field: GraphQLField):
    query = GraphQLObjectType(name="Query", fields={"singleField": field})
    return GraphQLSchema(query=query)


def describe_type_system_printer():
    def prints_string_field():
        schema = build_single_field_schema(GraphQLField(GraphQLString))
        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField: String
            }
            """
        )

    def prints_list_of_string_field():
        schema = build_single_field_schema(GraphQLField(GraphQLList(GraphQLString)))
        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField: [String]
            }
            """
        )

    def prints_non_null_string_field():
        schema = build_single_field_schema(GraphQLField(GraphQLNonNull(GraphQLString)))
        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField: String!
            }
            """
        )

    def prints_non_null_list_of_string_field():
        schema = build_single_field_schema(
            GraphQLField(GraphQLNonNull(GraphQLList(GraphQLString)))
        )
        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField: [String]!
            }
            """
        )

    def prints_list_of_non_null_string_field():
        schema = build_single_field_schema(
            GraphQLField((GraphQLList(GraphQLNonNull(GraphQLString))))
        )
        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField: [String!]
            }
            """
        )

    def prints_non_null_list_of_non_null_string_field():
        schema = build_single_field_schema(
            GraphQLField(GraphQLNonNull(GraphQLList(GraphQLNonNull(GraphQLString))))
        )
        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField: [String!]!
            }
            """
        )

    def prints_object_field():
        foo_type = GraphQLObjectType(
            name="Foo", fields={"str": GraphQLField(GraphQLString)}
        )
        schema = GraphQLSchema(types=[foo_type])

        assert expect_printed_schema(schema) == dedent(
            """
            type Foo {
              str: String
            }
            """
        )

    def prints_string_field_with_int_arg():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString, args={"argOne": GraphQLArgument(GraphQLInt)}
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int): String
            }
            """
        )

    def prints_string_field_with_int_arg_with_default():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={"argOne": GraphQLArgument(GraphQLInt, default_value=2)},
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int = 2): String
            }
            """
        )

    def prints_string_field_with_string_arg_with_default():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={
                    "argOne": GraphQLArgument(
                        GraphQLString, default_value="tes\t de\fault"
                    )
                },
            )
        )

        assert expect_printed_schema(schema) == dedent(
            r"""
            type Query {
              singleField(argOne: String = "tes\t de\fault"): String
            }
            """
        )

    def prints_string_field_with_int_arg_with_default_null():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={"argOne": GraphQLArgument(GraphQLInt, default_value=None)},
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int = null): String
            }
            """
        )

    def prints_string_field_with_non_null_int_arg():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={"argOne": GraphQLArgument(GraphQLNonNull(GraphQLInt))},
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int!): String
            }
            """
        )

    def prints_string_field_with_multiple_args():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={
                    "argOne": GraphQLArgument(GraphQLInt),
                    "argTwo": GraphQLArgument(GraphQLString),
                },
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int, argTwo: String): String
            }
            """
        )

    def prints_string_field_with_multiple_args_first_is_default():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={
                    "argOne": GraphQLArgument(GraphQLInt, default_value=1),
                    "argTwo": GraphQLArgument(GraphQLString),
                    "argThree": GraphQLArgument(GraphQLBoolean),
                },
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int = 1, argTwo: String, argThree: Boolean): String
            }
            """
        )

    def prints_string_field_with_multiple_args_second_is_default():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={
                    "argOne": GraphQLArgument(GraphQLInt),
                    "argTwo": GraphQLArgument(GraphQLString, default_value="foo"),
                    "argThree": GraphQLArgument(GraphQLBoolean),
                },
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int, argTwo: String = "foo", argThree: Boolean): String
            }
            """  # noqa: E501
        )

    def prints_string_field_with_multiple_args_last_is_default():
        schema = build_single_field_schema(
            GraphQLField(
                type_=GraphQLString,
                args={
                    "argOne": GraphQLArgument(GraphQLInt),
                    "argTwo": GraphQLArgument(GraphQLString),
                    "argThree": GraphQLArgument(GraphQLBoolean, default_value=False),
                },
            )
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              singleField(argOne: Int, argTwo: String, argThree: Boolean = false): String
            }
            """  # noqa: E501
        )

    def prints_schema_with_description():
        schema = GraphQLSchema(
            description="Schema description.", query=GraphQLObjectType("Query", {})
        )

        assert expect_printed_schema(schema) == dedent(
            '''
            """Schema description."""
            schema {
              query: Query
            }

            type Query
            '''
        )

    def omits_schema_of_common_names():
        schema = GraphQLSchema(
            query=GraphQLObjectType("Query", {}),
            mutation=GraphQLObjectType("Mutation", {}),
            subscription=GraphQLObjectType("Subscription", {}),
        )

        assert expect_printed_schema(schema) == dedent(
            """
            type Query

            type Mutation

            type Subscription
            """
        )

    def prints_custom_query_root_types():
        schema = GraphQLSchema(query=GraphQLObjectType("CustomType", {}))

        assert expect_printed_schema(schema) == dedent(
            """
            schema {
              query: CustomType
            }

            type CustomType
            """
        )

    def prints_custom_mutation_root_types():
        schema = GraphQLSchema(mutation=GraphQLObjectType("CustomType", {}))

        assert expect_printed_schema(schema) == dedent(
            """
            schema {
              mutation: CustomType
            }

            type CustomType
            """
        )

    def prints_custom_subscription_root_types():
        schema = GraphQLSchema(subscription=GraphQLObjectType("CustomType", {}))

        assert expect_printed_schema(schema) == dedent(
            """
            schema {
              subscription: CustomType
            }

            type CustomType
            """
        )

    def prints_interface():
        foo_type = GraphQLInterfaceType(
            name="Foo", fields={"str": GraphQLField(GraphQLString)}
        )

        bar_type = GraphQLObjectType(
            name="Bar",
            fields={"str": GraphQLField(GraphQLString)},
            interfaces=[foo_type],
        )

        schema = GraphQLSchema(types=[bar_type])
        assert expect_printed_schema(schema) == dedent(
            """
            type Bar implements Foo {
              str: String
            }

            interface Foo {
              str: String
            }
            """
        )

    def prints_multiple_interfaces():
        foo_type = GraphQLInterfaceType(
            name="Foo", fields={"str": GraphQLField(GraphQLString)}
        )

        baz_type = GraphQLInterfaceType(
            name="Baz", fields={"int": GraphQLField(GraphQLInt)}
        )

        bar_type = GraphQLObjectType(
            name="Bar",
            fields={
                "str": GraphQLField(GraphQLString),
                "int": GraphQLField(GraphQLInt),
            },
            interfaces=[foo_type, baz_type],
        )

        schema = GraphQLSchema(types=[bar_type])
        assert expect_printed_schema(schema) == dedent(
            """
            type Bar implements Foo & Baz {
              str: String
              int: Int
            }

            interface Foo {
              str: String
            }

            interface Baz {
              int: Int
            }
            """
        )

    def prints_hierarchical_interface():
        foo_type = GraphQLInterfaceType(
            name="Foo", fields={"str": GraphQLField(GraphQLString)}
        )

        baz_type = GraphQLInterfaceType(
            name="Baz",
            interfaces=[foo_type],
            fields={
                "int": GraphQLField(GraphQLInt),
                "str": GraphQLField(GraphQLString),
            },
        )

        bar_type = GraphQLObjectType(
            name="Bar",
            fields={
                "str": GraphQLField(GraphQLString),
                "int": GraphQLField(GraphQLInt),
            },
            interfaces=[foo_type, baz_type],
        )

        query = GraphQLObjectType(name="Query", fields={"bar": GraphQLField(bar_type)})

        schema = GraphQLSchema(query, types=[bar_type])
        assert expect_printed_schema(schema) == dedent(
            """
            type Bar implements Foo & Baz {
              str: String
              int: Int
            }

            interface Foo {
              str: String
            }

            interface Baz implements Foo {
              int: Int
              str: String
            }

            type Query {
              bar: Bar
            }
            """
        )

    def prints_unions():
        foo_type = GraphQLObjectType(
            name="Foo", fields={"bool": GraphQLField(GraphQLBoolean)}
        )

        bar_type = GraphQLObjectType(
            name="Bar", fields={"str": GraphQLField(GraphQLString)}
        )

        single_union = GraphQLUnionType(name="SingleUnion", types=[foo_type])

        multiple_union = GraphQLUnionType(
            name="MultipleUnion", types=[foo_type, bar_type]
        )

        schema = GraphQLSchema(types=[single_union, multiple_union])
        assert expect_printed_schema(schema) == dedent(
            """
            union SingleUnion = Foo

            type Foo {
              bool: Boolean
            }

            union MultipleUnion = Foo | Bar

            type Bar {
              str: String
            }
            """
        )

    def prints_input_type():
        input_type = GraphQLInputObjectType(
            name="InputType", fields={"int": GraphQLInputField(GraphQLInt)}
        )

        schema = GraphQLSchema(types=[input_type])
        assert expect_printed_schema(schema) == dedent(
            """
            input InputType {
              int: Int
            }
            """
        )

    def prints_custom_scalar():
        odd_type = GraphQLScalarType(name="Odd")

        schema = GraphQLSchema(types=[odd_type])
        assert expect_printed_schema(schema) == dedent(
            """
            scalar Odd
            """
        )

    def prints_custom_scalar_with_specified_by_url():
        foo_type = GraphQLScalarType(
            name="Foo", specified_by_url="https://example.com/foo_spec"
        )

        schema = GraphQLSchema(types=[foo_type])
        assert expect_printed_schema(schema) == dedent(
            """
            scalar Foo @specifiedBy(url: "https://example.com/foo_spec")
            """
        )

    def prints_enum():
        rgb_type = GraphQLEnumType(
            name="RGB", values=dict.fromkeys(("RED", "GREEN", "BLUE"))
        )

        schema = GraphQLSchema(types=[rgb_type])
        assert expect_printed_schema(schema) == dedent(
            """
            enum RGB {
              RED
              GREEN
              BLUE
            }
            """
        )

    def prints_empty_types():
        schema = GraphQLSchema(
            types=[
                GraphQLEnumType("SomeEnum", cast(Dict[str, Any], {})),
                GraphQLInputObjectType("SomeInputObject", {}),
                GraphQLInterfaceType("SomeInterface", {}),
                GraphQLObjectType("SomeObject", {}),
                GraphQLUnionType("SomeUnion", []),
            ]
        )

        assert expect_printed_schema(schema) == dedent(
            """
            enum SomeEnum

            input SomeInputObject

            interface SomeInterface

            type SomeObject

            union SomeUnion
            """
        )

    def prints_custom_directives():
        simple_directive = GraphQLDirective(
            "simpleDirective", [DirectiveLocation.FIELD]
        )
        complex_directive = GraphQLDirective(
            "complexDirective",
            [DirectiveLocation.FIELD, DirectiveLocation.QUERY],
            description="Complex Directive",
            args={
                "stringArg": GraphQLArgument(GraphQLString),
                "intArg": GraphQLArgument(GraphQLInt, default_value=-1),
            },
            is_repeatable=True,
        )

        schema = GraphQLSchema(directives=[simple_directive, complex_directive])
        assert expect_printed_schema(schema) == dedent(
            '''
            directive @simpleDirective on FIELD

            """Complex Directive"""
            directive @complexDirective(stringArg: String, intArg: Int = -1) repeatable on FIELD | QUERY
            '''  # noqa: E501
        )

    def prints_an_empty_description():
        schema = build_single_field_schema(GraphQLField(GraphQLString, description=""))

        assert expect_printed_schema(schema) == dedent(
            '''
            type Query {
              """"""
              singleField: String
            }
            '''
        )

    def prints_a_description_with_only_whitespace():
        schema = build_single_field_schema(GraphQLField(GraphQLString, description=" "))

        assert expect_printed_schema(schema) == dedent(
            """
            type Query {
              " "
              singleField: String
            }
            """
        )

    def one_line_prints_a_short_description():
        schema = build_single_field_schema(
            GraphQLField(GraphQLString, description="This field is awesome")
        )

        assert expect_printed_schema(schema) == dedent(
            '''
            type Query {
              """This field is awesome"""
              singleField: String
            }
            '''
        )

    def prints_introspection_schema():
        schema = GraphQLSchema()
        output = print_introspection_schema(schema)

        assert output == dedent(
            '''
            """
            Directs the executor to include this field or fragment only when the `if` argument is true.
            """
            directive @include(
              """Included when true."""
              if: Boolean!
            ) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

            """
            Directs the executor to skip this field or fragment when the `if` argument is true.
            """
            directive @skip(
              """Skipped when true."""
              if: Boolean!
            ) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

            """Marks an element of a GraphQL schema as no longer supported."""
            directive @deprecated(
              """
              Explains why this element was deprecated, usually also including a suggestion for how to access supported similar data. Formatted using the Markdown syntax, as specified by [CommonMark](https://commonmark.org/).
              """
              reason: String = "No longer supported"
            ) on FIELD_DEFINITION | ARGUMENT_DEFINITION | INPUT_FIELD_DEFINITION | ENUM_VALUE

            """Exposes a URL that specifies the behaviour of this scalar."""
            directive @specifiedBy(
              """The URL that specifies the behaviour of this scalar."""
              url: String!
            ) on SCALAR

            """
            A GraphQL Schema defines the capabilities of a GraphQL server. It exposes all available types and directives on the server, as well as the entry points for query, mutation, and subscription operations.
            """
            type __Schema {
              description: String

              """A list of all types supported by this server."""
              types: [__Type!]!

              """The type that query operations will be rooted at."""
              queryType: __Type!

              """
              If this server supports mutation, the type that mutation operations will be rooted at.
              """
              mutationType: __Type

              """
              If this server support subscription, the type that subscription operations will be rooted at.
              """
              subscriptionType: __Type

              """A list of all directives supported by this server."""
              directives: [__Directive!]!
            }

            """
            The fundamental unit of any GraphQL Schema is the type. There are many kinds of types in GraphQL as represented by the `__TypeKind` enum.

            Depending on the kind of a type, certain fields describe information about that type. Scalar types provide no information beyond a name, description and optional `specifiedByURL`, while Enum types provide their values. Object and Interface types provide the fields they describe. Abstract types, Union and Interface, provide the Object types possible at runtime. List and NonNull types compose other types.
            """
            type __Type {
              kind: __TypeKind!
              name: String
              description: String
              specifiedByURL: String
              fields(includeDeprecated: Boolean = false): [__Field!]
              interfaces: [__Type!]
              possibleTypes: [__Type!]
              enumValues(includeDeprecated: Boolean = false): [__EnumValue!]
              inputFields(includeDeprecated: Boolean = false): [__InputValue!]
              ofType: __Type
            }

            """An enum describing what kind of type a given `__Type` is."""
            enum __TypeKind {
              """Indicates this type is a scalar."""
              SCALAR

              """
              Indicates this type is an object. `fields` and `interfaces` are valid fields.
              """
              OBJECT

              """
              Indicates this type is an interface. `fields`, `interfaces`, and `possibleTypes` are valid fields.
              """
              INTERFACE

              """Indicates this type is a union. `possibleTypes` is a valid field."""
              UNION

              """Indicates this type is an enum. `enumValues` is a valid field."""
              ENUM

              """
              Indicates this type is an input object. `inputFields` is a valid field.
              """
              INPUT_OBJECT

              """Indicates this type is a list. `ofType` is a valid field."""
              LIST

              """Indicates this type is a non-null. `ofType` is a valid field."""
              NON_NULL
            }

            """
            Object and Interface types are described by a list of Fields, each of which has a name, potentially a list of arguments, and a return type.
            """
            type __Field {
              name: String!
              description: String
              args(includeDeprecated: Boolean = false): [__InputValue!]!
              type: __Type!
              isDeprecated: Boolean!
              deprecationReason: String
            }

            """
            Arguments provided to Fields or Directives and the input fields of an InputObject are represented as Input Values which describe their type and optionally a default value.
            """
            type __InputValue {
              name: String!
              description: String
              type: __Type!

              """
              A GraphQL-formatted string representing the default value for this input value.
              """
              defaultValue: String
              isDeprecated: Boolean!
              deprecationReason: String
            }

            """
            One possible value for a given Enum. Enum values are unique values, not a placeholder for a string or numeric value. However an Enum value is returned in a JSON response as a string.
            """
            type __EnumValue {
              name: String!
              description: String
              isDeprecated: Boolean!
              deprecationReason: String
            }

            """
            A Directive provides a way to describe alternate runtime execution and type validation behavior in a GraphQL document.

            In some cases, you need to provide options to alter GraphQL's execution behavior in ways field arguments will not suffice, such as conditionally including or skipping a field. Directives provide this by describing additional information to the executor.
            """
            type __Directive {
              name: String!
              description: String
              isRepeatable: Boolean!
              locations: [__DirectiveLocation!]!
              args(includeDeprecated: Boolean = false): [__InputValue!]!
            }

            """
            A Directive can be adjacent to many parts of the GraphQL language, a __DirectiveLocation describes one such possible adjacencies.
            """
            enum __DirectiveLocation {
              """Location adjacent to a query operation."""
              QUERY

              """Location adjacent to a mutation operation."""
              MUTATION

              """Location adjacent to a subscription operation."""
              SUBSCRIPTION

              """Location adjacent to a field."""
              FIELD

              """Location adjacent to a fragment definition."""
              FRAGMENT_DEFINITION

              """Location adjacent to a fragment spread."""
              FRAGMENT_SPREAD

              """Location adjacent to an inline fragment."""
              INLINE_FRAGMENT

              """Location adjacent to a variable definition."""
              VARIABLE_DEFINITION

              """Location adjacent to a schema definition."""
              SCHEMA

              """Location adjacent to a scalar definition."""
              SCALAR

              """Location adjacent to an object type definition."""
              OBJECT

              """Location adjacent to a field definition."""
              FIELD_DEFINITION

              """Location adjacent to an argument definition."""
              ARGUMENT_DEFINITION

              """Location adjacent to an interface definition."""
              INTERFACE

              """Location adjacent to a union definition."""
              UNION

              """Location adjacent to an enum definition."""
              ENUM

              """Location adjacent to an enum value definition."""
              ENUM_VALUE

              """Location adjacent to an input object type definition."""
              INPUT_OBJECT

              """Location adjacent to an input object field definition."""
              INPUT_FIELD_DEFINITION
            }
            '''  # noqa: E501
        )


def describe_print_value():
    def print_value_convenience_function():
        assert print_value(1.5, GraphQLFloat) == "1.5"
        assert print_value("foo", GraphQLString) == '"foo"'



from graphql import parse
from graphql.language import visit, Visitor, ParallelVisitor

from ..fixtures import big_schema_sdl  # noqa: F401


class DummyVisitor(Visitor):
    @staticmethod
    def enter(*args):
        pass

    @staticmethod
    def leave(*args):
        pass


def test_visit_all_ast_nodes(benchmark, big_schema_sdl):  # noqa: F811
    document_ast = parse(big_schema_sdl)
    visitor = DummyVisitor()
    benchmark(lambda: visit(document_ast, visitor))


def test_visit_all_ast_nodes_in_parallel(benchmark, big_schema_sdl):  # noqa: F811
    document_ast = parse(big_schema_sdl)
    visitor = DummyVisitor()
    parallel_visitor = ParallelVisitor([visitor] * 50)
    benchmark(lambda: visit(document_ast, parallel_visitor))



from pytest import raises

from graphql.language import parse_type, TypeNode
from graphql.type import GraphQLList, GraphQLNonNull, GraphQLObjectType
from graphql.utilities import type_from_ast

from ..validation.harness import test_schema


def describe_type_from_ast():
    def for_named_type_node():
        node = parse_type("Cat")
        type_for_node = type_from_ast(test_schema, node)
        assert isinstance(type_for_node, GraphQLObjectType)
        assert type_for_node.name == "Cat"

    def for_list_type_node():
        node = parse_type("[Cat]")
        type_for_node = type_from_ast(test_schema, node)
        assert isinstance(type_for_node, GraphQLList)
        of_type = type_for_node.of_type
        assert isinstance(of_type, GraphQLObjectType)
        assert of_type.name == "Cat"

    def for_non_null_type_node():
        node = parse_type("Cat!")
        type_for_node = type_from_ast(test_schema, node)
        assert isinstance(type_for_node, GraphQLNonNull)
        of_type = type_for_node.of_type
        assert isinstance(of_type, GraphQLObjectType)
        assert of_type.name == "Cat"

    def for_unspecified_type_node():
        node = TypeNode()
        with raises(TypeError) as exc_info:
            type_from_ast(test_schema, node)
        msg = str(exc_info.value)
        assert msg == "Unexpected type node: <TypeNode instance>."
